import math
import numpy as np
import torch
from typing import Optional

from torchness.base_elements import TorchnessException
from torchness.types import ACT, INI, TNS
from torchness.base_elements import my_initializer


# my dense layer, linear with initializer + activation
class LayDense(torch.nn.Linear):

    def __init__(
            self,
            in_features: int,
            out_features: int,
            activation: ACT=    torch.nn.ReLU,
            bias: bool=         True,
            device=             None,
            dtype=              None,
            initializer: INI=   None):
        self.initializer = initializer or my_initializer
        torch.nn.Linear.__init__(
            self,
            in_features=    in_features,
            out_features=   out_features,
            bias=           bias,
            device=         device,
            dtype=          dtype)
        self.activation = activation() if activation else None

    def reset_parameters(self) -> None:

        self.initializer(self.weight)
        if self.bias is not None:
            torch.nn.init.zeros_(self.bias)

            ### original Linear (with uniform) reset for bias
            # fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight)
            # bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            # torch.nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, inp:TNS) -> TNS:
        out = super().forward(inp)
        if self.activation: out = self.activation(out)
        return out

    def extra_repr(self) -> str:
        act_info = '' if self.activation else ', activation=None'
        return f'in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}{act_info}'

# time & feats dropout (for sequences), inp tensor [..,seq,feats]
class TF_Dropout(torch.nn.Dropout):

    def __init__(
            self,
            time_drop: float=   0.0,
            feat_drop: float=   0.0,
            inplace: bool=      False):
        self.time_drop = time_drop
        self.feat_drop = feat_drop
        super(TF_Dropout, self).__init__(inplace=inplace)

    def forward(self, inp:TNS) -> TNS:

        output = inp
        in_shape = inp.size()

        if self.time_drop:
            t_drop = torch.ones(in_shape[-2])
            t_drop = torch.nn.functional.dropout(
                input=      t_drop,
                p=          self.time_drop,
                training=   self.training,
                inplace=    self.inplace)
            t_drop = torch.unsqueeze(t_drop, dim=-1)
            output = output * t_drop

        if self.feat_drop:
            f_drop = torch.ones(in_shape[-1])
            f_drop = torch.nn.functional.dropout(
                input=      f_drop,
                p=          self.feat_drop,
                training=   self.training,
                inplace=    self.inplace)
            f_drop = torch.unsqueeze(f_drop, dim=-2)
            output = output * f_drop

        return output

    def extra_repr(self) -> str:
        return f'time_drop={self.time_drop}, feat_drop={self.feat_drop}, inplace={self.inplace}'

# my Conv1D, with initializer + activation
class LayConv1D(torch.nn.Conv1d):

    def __init__(
            self,
            in_features: int,                   # input num of channels
            n_filters: int,                     # output num of channels
            kernel_size: int=   3,
            stride=             1,              # single number or a one-element tuple
            padding=            'same',
            dilation=           1,
            groups=             1,
            bias=               True,
            padding_mode=       'zeros',
            device=             None,
            dtype=              None,
            activation: ACT=    torch.nn.ReLU,
            initializer: INI=   None):

        super(LayConv1D, self).__init__(
            in_channels=    in_features,
            out_channels=   n_filters,
            kernel_size=    kernel_size,
            stride=         stride,
            padding=        padding,
            dilation=       dilation,
            groups=         groups,
            bias=           bias,
            padding_mode=   padding_mode,
            device=         device,
            dtype=          dtype)

        self.activation = activation() if activation else None

        if not initializer: initializer = my_initializer
        initializer(self.weight)
        if self.bias is not None:
            torch.nn.init.zeros_(self.bias)

            # original Conv1D (with uniform) reset for bias
            # fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight)
            # bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            # torch.nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, inp:TNS) -> TNS:
        inp_trans = torch.transpose(input=inp, dim0=-1, dim1=-2) # transposes inp to (N,C,L) <- (N,L,C), since torch.nn.Conv1d assumes that channels is @ -2 dim
        out = super().forward(input=inp_trans)
        out = torch.transpose(out, dim0=-1, dim1=-2) # transpose back
        if self.activation: out = self.activation(out)
        return out

# Residual Layer with dropout for bypass inp
class LayRES(torch.nn.Module):

    def __init__(
            self,
            in_features: Optional[int]= None,
            dropout: float=             0.0):

        if dropout and in_features is None:
            raise TorchnessException('LayRES with dropout needs to know its in_features (int) - cannot be None')

        super(LayRES, self).__init__()

        self.dropout = torch.nn.Dropout(p=dropout) if dropout else None

    def forward(self, inp:TNS, bypass:TNS) -> TNS:
        if self.dropout:
            bypass = self.dropout(bypass)
        return inp + bypass


class PositionalEncoding(torch.nn.Module):

    def __init__(
            self,
            d_model: int,
            dropout: float= 0.0,
            max_len: int=   512):

        super().__init__()

        self.dropout = torch.nn.Dropout(p=dropout) if dropout else None

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(max_len, d_model)
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    # x - tensor of shape [..,seq,feats]
    def forward(self, x:TNS) -> TNS:
        x = x + self.pe[:x.size(-2)]
        if self.dropout:
            x = self.dropout(x)
        return x


# returns [0,1] tensor: 1 where inp not activated (value =< 0), looks at last dimension / features
def zeroes(inp:TNS) -> np.ndarray:
    activated = (inp > 0).to(int)
    axes = list(range(len(inp.shape)))[:-1]  # all but last(feats) axes indexes list like: [0,1,2] for 4d shape
    activated_reduced = torch.sum(activated, dim=axes) if axes else activated  # 1 or more for activated, 0 for not activated, if not axes -> we have only-feats-tensor-case
    return (activated_reduced == 0).to(int).detach().cpu().numpy() # to ndarray


import torch
import unittest

from torchness.tbwr import TBwr

from tests.envy import flush_tmp_dir

TBWR_DIR = f'{flush_tmp_dir()}/tbwr'


class TestTBwr(unittest.TestCase):

    def setUp(self) -> None:
        flush_tmp_dir()

    def test_TBwr_values(self):

        tbwr = TBwr(logdir=f'{TBWR_DIR}/val')

        val = 1.7
        for ix in range(100):
            tbwr.add(value=val, tag='val', step=ix)
            val += 0.15

    def test_TBwr_histogram(self):

        tbwr = TBwr(logdir=f'{TBWR_DIR}/values_histogram')

        vals = torch.rand(100)
        for ix in range(100):
            tbwr.add_histogram(values=vals, tag='vals_hist', step=ix)
            vals = vals/1.05 + torch.rand(100)*0.05




"""
    MOTorch wraps PyTorch neural network (Module) and adds some features.
    - Builds given Module.
    - Manages one MOTorch folder (subfolder of SAVE_TOPDIR named with MOTorch name)
      for all MOTorch data (logs, params, checkpoints). MOTorch supports
      serialization into this folder.
    - Extends ParaSave, manages all init parameters. Properly resolves parameters
      using all possible sources, saves and loads them from MOTorch folder.
    - Parameters are kept in self as a Subscriptable to be easily accessed.
    - Properly resolves and holds name of object, adds stamp if needed.
    - Supports / creates logger.
    - May be read only (prevents save over).
    - May be called (with __call__) <- runs NN FWD with given data
    - May be called BWD with backward() <- runs gradient backprop for given data
    - Supports hpmser mode.
    - Manages seed and guarantees reproducibility.
    - Manages GPU / CPU device used by NN.
    - Adds TensorBoard support.
    - Defines / implements save / load / copy of whole MOTorch (ParaSave + NN checkpoint).
    - Defines interface of baseline training & testing with data loaded to Batcher.
    - Defines / implements GX.
    - Adds some sanity checks.

    Module:
        - should implement forward() and loss() methods
            - arguments for parameters of forward() will be cast to TNS (by default) by MOTorch,
        - device is managed by MOTorch with dev_pypaq: DevicesPypaq parameter

    MOTorch extends Module. By default, after init, MOTorch is set to train.mode=False.
    MOTorch manages its train.mode by itself.
"""

import numpy as np
import shutil
from sklearn.metrics import f1_score
import torch
from typing import Optional, Dict, Tuple, Any

from pypaq.lipytools.printout import stamp
from pypaq.lipytools.files import prep_folder
from pypaq.lipytools.pylogger import get_pylogger, get_child
from pypaq.lipytools.moving_average import MovAvg
from pypaq.pms.base import get_class_init_params, point_trim
from pypaq.pms.parasave import ParaSave
from pypaq.mpython.devices import get_devices
from pypaq.mpython.mpdecor import proc_wait
from torchness.comoneural.batcher import Batcher
from torchness.types import TNS, DTNS
from torchness.base_elements import mrg_ckpts
from torchness.scaled_LR import ScaledLR
from torchness.grad_clipping import GradClipperMAVG
from torchness.tbwr import TBwr



class MOTorchException(Exception):
    pass


# torch.nn.Module to be implemented
# forward() is needed by torch.nn.Module, loss() is needed by MOTorch: backward() & run_train()
class Module(torch.nn.Module):

    def __init__(
            self,
            logger=     None,
            loglevel=   20,
            **kwargs):
        torch.nn.Module.__init__(self)
        if not logger:
            logger = get_pylogger(name='Module_logger', level=loglevel)
        self.logger = logger

    # returned DTNS should have at least 'logits' key with logits tensor for proper MOTorch.run_train()
    def forward(self, *args, **kwargs) -> DTNS:
        """
            exemplary implementation:
        return {'logits': self.logits(input)}
        """
        raise NotImplementedError

    # baseline accuracy implementation for logits & lables
    def accuracy(
            self,
            logits: TNS,
            labels: TNS) -> float:
        logits = logits.detach().cpu().numpy()
        preds = np.argmax(logits, axis=-1)
        labels = labels.cpu().numpy()
        return float(np.average(np.equal(preds, labels)))

    # baseline F1 implementation for logits & lables
    # correctly supports average (since test while training may be run in batches):
    #  micro (per sample)
    #  macro (per class)
    #  weighted (per class weighted by support)
    def f1(self, logits:TNS, labels:TNS, average='weighted') -> float:
        logits = logits.detach().cpu().numpy()
        preds = np.argmax(logits, axis=-1)
        labels = labels.cpu().numpy()
        return f1_score(
            y_true=         labels,
            y_pred=         preds,
            average=        average,
            labels=         np.unique(preds),
            zero_division=  0)

    # returned DTNS should be: forward() DTNS updated with loss (and optional acc, f1)
    def loss(self, *args, **kwargs) -> DTNS:
        """
            exemplary implementation:
        out = self(input)                                                                   <- forward DTNS
        logits = out['logits']
        out['loss'] = torch.nn.functional.cross_entropy(logits, labels, reduction='mean')   <- update with loss
        out['acc'] = self.accuracy(logits, labels)                                          <- update with acc
        out['f1'] = self.f1(logits, labels)                                                 <- update with f1
        """
        raise NotImplementedError


# extends Module (torch.nn.Module) with ParaSave and many others
class MOTorch(ParaSave, torch.nn.Module):

    SPEC_KEYS: set = {
        'loss',         # loss
        'acc',          # accuracy
        'f1'}           # F1

    MOTORCH_DEFAULTS = {
        'seed':             123,                # seed for torch and numpy
        'device':           -1,                 # :DevicesPypaq (check pypaq.mpython.devices)
        'dtype':            torch.float32,      # dtype of floats in MOTorch (16/32/64 etc)
        'bypass_data_conv': False,              # to bypass input data conversion with forward(), loss(), backward()
            # training
        'batch_size':       64,                 # training batch size
        'n_batches':        1000,               # default length of training
        'opt_class':        torch.optim.Adam,   # default optimizer
        'train_step':       0,                  # default (starting) train step, updated with backward()
            # LR management (check torchness.base_elements.ScaledLR)
        'baseLR':           3e-4,
        'warm_up':          None,
        'ann_base':         None,
        'ann_step':         1.0,
        'n_wup_off':        2.0,
            # gradients clipping parameters (check torchness.base_elements.GradClipperMAVG)
        'gc_factor':        0.01,
        'gc_first_avg':     True,
        'gc_start_val':     0.1,
        'gc_max_upd':       1.5,
        'do_clip':          False,
            # other
        'try_load_ckpt':    True,               # tries to load a checkpoint while init
        'hpmser_mode':      False,              # it will set model to be read_only and quiet when running with hpmser
        'read_only':        False,              # sets MOTorch to be read only - won't save anything (won't even create self.motorch_dir)
        'do_TB':            True,               # runs TensorBard, saves in self.motorch_dir
    }

    SAVE_TOPDIR = '_models'
    SAVE_FN_PFX = 'motorch_point' # POINT file prefix

    def __init__(
            self,
            module_type: Optional[type(Module)]=    None,   # also accepts torch.nn.Module but then some methods won't work (run_train, etc.)
            name: Optional[str]=                    None,
            name_timestamp=                         False,
            save_topdir: Optional[str]=             None,
            save_fn_pfx: Optional[str]=             None,
            tbwr: Optional[TBwr]=                   None,
            logger=                                 None,
            loglevel=                               20,
            flat_child=                             False,
            **kwargs):

        if not name and not module_type:
            raise MOTorchException('name OR module_type must be given!')

        # TODO: temporary, delete later
        if 'devices' in kwargs:
            raise MOTorchException('\'devices\' param is no more supported by MOTorch, please use \'device\'')

        self.module_type = module_type

        # generate name
        name = f'{self.module_type.__name__}_MOTorch' if not name else name
        if name_timestamp: name += f'_{stamp()}'
        self.name = name

        # some early overrides

        if kwargs.get('hpmser_mode', False):
            loglevel = 50
            kwargs['read_only'] = True

        if kwargs.get('read_only', False):
            kwargs['do_TB'] = False

        _read_only = kwargs.get('read_only', False)

        if not save_topdir: save_topdir = self.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = self.SAVE_FN_PFX

        if not logger:
            logger = get_pylogger(
                name=       self.name,
                add_stamp=  False,
                folder=     None if _read_only else MOTorch.__get_model_dir(save_topdir, self.name),
                level=      loglevel,
                flat_child= flat_child)
        self._log = logger

        mod_info = self.module_type.__name__ if self.module_type else 'module_type NOT GIVEN (will try to load from saved)'
        self._log.info(f'*** MOTorch : {self.name} *** initializes for module_type: {mod_info}')
        self._log.info(f'> {self.name} save_topdir: {save_topdir}{" <- read only mode!" if _read_only else ""}')

        # ************************************************************************************************* manage point

        # try to load point from given folder
        point_saved = ParaSave.load_point(
            name=           self.name,
            save_topdir=    save_topdir,
            save_fn_pfx=    save_fn_pfx)

        ### resolve module_type

        # if 'module_type' was not given with init -> try to get it from saved
        if not self.module_type:
            self.module_type = point_saved.get('module_type', None)
        if not self.module_type:
            msg = 'module_type was not given and has not been found in saved, cannot continue!'
            self._log.error(msg)
            raise MOTorchException(msg)

        ### manage params from self.module_type.__init__

        _module_init_def = get_class_init_params(self.module_type)['with_defaults'] # defaults of self.module_type.__init__

        # special case of params: [device, dtype] <- those will be set with values prepared by MOTorch below, BUT...
        _override_in_module_for_none = {
            'device':   MOTorch.MOTORCH_DEFAULTS['device'],
            'dtype':    MOTorch.MOTORCH_DEFAULTS['dtype']}

        # ..EXCEPT a case when are set in self.module_type.__init__ to other value than None
        remove_from_override = []
        for param in _override_in_module_for_none:
            if param in _module_init_def and _module_init_def[param] is not None:
                remove_from_override.append(param)
        for param in remove_from_override:
            _override_in_module_for_none.pop(param)

        ### update in proper order

        self._point = {}
        self._point.update(ParaSave.PARASAVE_DEFAULTS)
        self._point.update(MOTorch.MOTORCH_DEFAULTS)
        self._point.update(_module_init_def)
        self._point.update(_override_in_module_for_none)
        self._point.update(point_saved)
        self._point.update(kwargs)  # update with kwargs given NOW by user
        self._point['name'] = self.name
        self._point['save_topdir'] = save_topdir
        self._point['save_fn_pfx'] = save_fn_pfx

        # remove logger (may come from Module init defaults)
        if 'logger' in self._point:
            self._point.pop('logger')

        ### finally resolve device

        # device parameter, may be given to MOTorch in DevicesPypaq type - it needs to be cast to PyTorch namespace here
        self._log.debug(f'> {self.name} resolves devices, given: {self._point["device"]}')
        self._log.debug(f'>> torch.cuda.is_available(): {torch.cuda.is_available()}')
        devices = get_devices(
            devices=            self._point["device"],
            torch_namespace=    True,
            logger=             get_child(self._log, 'get_devices'))
        if not devices:
            self._log.warning(f'given device: {self._point["device"]} is not available, using CPU')
            devices = ['cpu']
        device = devices[0]
        self._log.info(f'> {self.name} given devices: {self._point["device"]}, will use: {device}')
        self._point['device'] = device

        ### prepare Module point and manage not used kwargs

        self._module_point = point_trim(self.module_type, self._point)
        self._module_point['logger'] = get_child(self._log, 'Module_logger')

        _kwargs_not_used = {}
        for k in kwargs:
            if k not in self._module_point:
                _kwargs_not_used[k] = kwargs[k]

        ### report

        self._log.debug(f'> {self.name} POINT sources:')
        self._log.debug(f'>> PARASAVE_DEFAULTS:         {ParaSave.PARASAVE_DEFAULTS}')
        self._log.debug(f'>> MOTORCH_DEFAULTS:          {MOTorch.MOTORCH_DEFAULTS}')
        self._log.debug(f'>> Module.__init__ defaults:  {_module_init_def}') # here are reported original Module.__init__ defaults without any MOTorch override
        self._log.debug(f'>> POINT saved:               {point_saved}')
        self._log.debug(f'>> given kwargs:              {kwargs}')
        self._log.debug(f'> resolved POINT:')
        self._log.debug(f'Module complete POINT:        {self._module_point}')
        self._log.debug(f'>> kwargs not used by Module: {_kwargs_not_used}')
        self._log.debug(f'{self.name} complete POINT:\n{self._point}')

        # ******************************************************************************************* init as a ParaSave

        ParaSave.__init__(self, logger=get_child(self._log, 'ParaSave_logger'), **self._point)

        # params names safety check
        pms = sorted(list(self.SPEC_KEYS) + list(MOTorch.MOTORCH_DEFAULTS.keys()) + list(kwargs.keys()))
        found = self.check_params_sim(params=pms)
        if found:
            self._log.warning('MOTorch was asked to check for params similarity and found:')
            for pa, pb in found: self._log.warning(f'> params \'{pa}\' and \'{pb}\' are close !!!')

        # *********************** set seed in all possible areas (https://pytorch.org/docs/stable/notes/randomness.html)

        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

        # ***************************************************************************************** build MOTorch Module

        self._log.info(f'{self.name} builds graph')
        torch.nn.Module.__init__(self) # init self as a torch.nn.Module
        self._module = self.module_type(**self._module_point)

        if self.try_load_ckpt:
            self.load_ckpt()
        else:
            self._log.info(f'> {self.name} checkpoint not loaded, not even tried because \'try_load_ckpt\' was set to {self.try_load_ckpt}')

        self._log.debug(f'> moving {self.name} to device: {self.device}, dtype: {self.dtype}')
        self.to(self.device)
        self.to(self.dtype)

        self._log.debug(f'{self.name} module initialized!')

        # optimizer params may be given with 'opt_' prefix
        opt_params = {k[4:]: self[k] for k in self.get_managed_params() if k.startswith('opt_')}
        opt_params.pop('class')
        self._opt = self.opt_class(
            params= self.parameters(),
            lr=     self.baseLR,
            **opt_params)
        #print(len(self._opt.param_groups))
        #print(self._opt.param_groups[0].keys())

        # from now LR is managed by scheduler
        self._scheduler = ScaledLR(
            optimizer=      self._opt,
            starting_step=  self.train_step,
            warm_up=        self.warm_up,
            ann_base=       self.ann_base,
            ann_step=       self.ann_step,
            n_wup_off=      self.n_wup_off,
            logger=         get_child(self._log, 'ScaledLR'))

        self._grad_clipper = GradClipperMAVG(
            module=         self,
            factor=         self.gc_factor,
            first_avg=      self.gc_first_avg,
            start_val=      self.gc_start_val,
            max_upd=        self.gc_max_upd,
            do_clip=        self.do_clip)

        # MOTorch by default is not in training mode
        self.train(False)
        self._log.debug(f'> set {self.name} train.mode to False..')

        # *********************************************************************************************** other & finish

        self._TBwr = tbwr or TBwr(logdir=MOTorch.__get_model_dir(self.save_topdir, self.name)) if self.do_TB else None  # TensorBoard writer

        self._batcher = None

        self._log.debug(str(self))
        self._log.info(f'MOTorch init finished!')

    # **************************************************************************** model call (run NN with data) methods

    # forward call on NN
    def __call__(self, *args, **kwargs) -> dict:
        return torch.nn.Module.__call__(self, *args, **kwargs)

    # converts given data to torch.Tensor compatible with self (type,device,dtype)
    def convert(self, data:Any) -> TNS:

        # do not convert None
        if type(data) is not None:

            if type(data) is not torch.Tensor:
                if type(data) is np.ndarray: data = torch.from_numpy(data)
                else:                        data = torch.tensor(data)

            # convert device + float types
            data = data.to(self.device, self.dtype if data.is_floating_point() or data.is_complex() else None)

        return data

    # runs forward on nn.Module (with current nn.Module.training.mode - by default not training)
    def forward(
            self,
            *args,
            bypass_data_conv=               False,
            set_training: Optional[bool]=   None,
            **kwargs) -> DTNS:
        """
        INFO: since MOTorch is a torch.nn.Module, call forward() call should be avoided,
        INFO: instead use just MOTorch.__call__() /self()
        """

        if set_training is not None: self.train(set_training)

        if not (bypass_data_conv or self.bypass_data_conv):
            args = [self.convert(data=a) for a in args]
            kwargs = {k: self.convert(data=kwargs[k]) for k in kwargs}

        out = self.module(*args, **kwargs)

        if set_training: self.train(False) # eventually roll back to default
        return out

    # forward + loss call on NN (with current nn.Module.training.mode - by default not training)
    def loss(
            self,
            *args,
            bypass_data_conv=               False,
            set_training: Optional[bool]=   None,   # for not None forces given training mode for torch.nn.Module
            **kwargs) -> DTNS:

        if set_training is not None: self.train(set_training)

        if not (bypass_data_conv or self.bypass_data_conv):
            args = [self.convert(data=a) for a in args]
            kwargs = {k: self.convert(data=kwargs[k]) for k in kwargs}

        out = self.module.loss(*args, **kwargs)

        if set_training: self.train(False) # eventually roll back to default
        return out

    # backward call on NN, runs loss calculation + update of nn.Module (by default with training.mode = True)
    def backward(
            self,
            *args,
            bypass_data_conv=   False,
            set_training: bool= True,   # for backward training mode is set to True by default
            empty_cuda_cache=   False,  # releases all unoccupied cached memory currently held by the caching allocator INFO: BEWARE: allocates another memory in cuda:0
            **kwargs) -> DTNS:

        out = self.loss(
            *args,
            bypass_data_conv=   bypass_data_conv,
            set_training=       set_training,
            **kwargs)

        out['loss'].backward()              # update gradients
        gnD = self._grad_clipper.clip()     # clip gradients, adds: 'gg_norm' & 'gg_norm_clip' to out
        self._opt.step()                    # apply optimizer
        self._opt.zero_grad()               # clear gradients
        self._scheduler.step()              # apply LR scheduler
        self.train_step += 1                # update step

        if empty_cuda_cache:
            torch.cuda.empty_cache()

        out['currentLR'] = self._scheduler.get_last_lr()[0]         # INFO: we take currentLR of first group
        out.update(gnD)

        return out

    # *********************************************************************************************** load / save / copy

    @staticmethod
    def __get_model_dir(save_topdir:str, model_name:str) -> str:
        return f'{save_topdir}/{model_name}'

    # returns path of checkpoint pickle file
    @staticmethod
    def __get_ckpt_path(save_topdir:str, model_name:str) -> str:
        return f'{MOTorch.__get_model_dir(save_topdir, model_name)}/{model_name}.pt'

    # tries to load checkpoint and return additional data
    def load_ckpt(
            self,
            save_topdir: Optional[str]= None,  # allows to load from custom save_topdir
            name: Optional[str]=        None,  # allows to load custom name (model_name)
    ) -> Optional[dict]:

        ckpt_path = MOTorch.__get_ckpt_path(
            save_topdir=    save_topdir or self.save_topdir,
            model_name=     name or self.name)

        save_obj = None

        try:
            # INFO: immediately place all tensors to current device (not previously saved one)
            save_obj = torch.load(f=ckpt_path, map_location=self.device)
            self.load_state_dict(save_obj.pop('model_state_dict'))
            self._log.info(f'> {self.name} checkpoint loaded from {ckpt_path}')
        except Exception as e:
            self._log.info(f'> {self.name} checkpoint NOT loaded because of exception: {e}')

        return save_obj

    # saves model checkpoint & optionally additional data
    def save_ckpt(
            self,
            save_topdir: Optional[str]=         None,   # allows to save in custom save_topdir
            name: Optional[str]=                None,   # allows to save under custom name (model_name)
            additional_data: Optional[Dict]=    None,   # allows to save additional
    ) -> None:

        ckpt_path = MOTorch.__get_ckpt_path(
            save_topdir=    save_topdir or self.save_topdir,
            model_name=     name or self.name)

        save_obj = {'model_state_dict': self.state_dict()}
        if additional_data: save_obj.update(additional_data)

        torch.save(obj=save_obj, f=ckpt_path)

    # saves MOTorch (ParaSave POINT and model checkpoint)
    def save(self):

        if self.read_only:
            raise MOTorchException('read_only MOTorch cannot be saved!')

        # to properly start grad clipping after load
        self['gc_first_avg'] = False
        self['gc_start_val'] = self._grad_clipper.gg_norm_clip

        self.save_point()
        self.save_ckpt()
        self._log.info(f'{self.__class__.__name__} {self.name} saved to {self.save_topdir}')

    @classmethod
    def copy_checkpoint(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None):
        if not save_topdir_src: save_topdir_src = cls.SAVE_TOPDIR
        if not save_topdir_trg: save_topdir_trg = save_topdir_src
        shutil.copyfile(
            src=    MOTorch.__get_ckpt_path(save_topdir_src, name_src),
            dst=    MOTorch.__get_ckpt_path(save_topdir_trg, name_trg))

    # copies full MOTorch folder (POINT & checkpoints)
    @classmethod
    def copy_saved(
            cls,
            name_src: str,
            name_trg: str,
            save_topdir_src: Optional[str]= None,
            save_topdir_trg: Optional[str]= None,
            save_fn_pfx: Optional[str]=     None,
            logger=                         None,
            loglevel=                       30):

        if not save_topdir_src: save_topdir_src = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        if save_topdir_trg is None: save_topdir_trg = save_topdir_src

        cls.copy_saved_point(
            name_src=           name_src,
            name_trg=           name_trg,
            save_topdir_src=    save_topdir_src,
            save_topdir_trg=    save_topdir_trg,
            save_fn_pfx=        save_fn_pfx,
            logger=             logger,
            loglevel=           loglevel)

        cls.copy_checkpoint(
            name_src=           name_src,
            name_trg=           name_trg,
            save_topdir_src=    save_topdir_src,
            save_topdir_trg=    save_topdir_trg)

    # *************************************************************************************************************** GX

    @classmethod
    def gx_ckpt(
            cls,
            name_A: str,                            # name parent A
            name_B: str,                            # name parent B
            name_child: str,                        # name child
            save_topdir_A: Optional[str]=       None,
            save_topdir_B: Optional[str]=       None,
            save_topdir_child: Optional[str]=   None,
            ratio: float=                       0.5,
            noise: float=                       0.03):

        if not save_topdir_A: save_topdir_A = cls.SAVE_TOPDIR
        if not save_topdir_B: save_topdir_B = save_topdir_A
        if not save_topdir_child: save_topdir_child = save_topdir_A

        prep_folder(f'{save_topdir_child}/{name_child}')

        mrg_ckpts(
            ckptA=          MOTorch.__get_ckpt_path(save_topdir_A, name_A),
            ckptB=          MOTorch.__get_ckpt_path(save_topdir_B, name_B),
            ckptM=          MOTorch.__get_ckpt_path(save_topdir_child, name_child),
            ratio=          ratio,
            noise=          noise)

    # performs GX on saved MOTorch (without even building child objects)
    @classmethod
    def gx_saved(
            cls,
            name_parent_main: str,
            name_parent_scnd: Optional[str],    # if not given makes GX only with main parent
            name_child: str,
            save_topdir_parent_main: Optional[str]= None,
            save_topdir_parent_scnd: Optional[str]= None,
            save_topdir_child: Optional[str]=       None,
            save_fn_pfx: Optional[str]=             None,
            do_gx_ckpt=                             True,
            ratio: float=                           0.5,
            noise: float=                           0.03,
            logger=                                 None,
            loglevel=                               30,
    ) -> None:

        if not save_topdir_parent_main: save_topdir_parent_main = cls.SAVE_TOPDIR
        if not save_fn_pfx: save_fn_pfx = cls.SAVE_FN_PFX

        cls.gx_saved_point(
            name_parent_main=           name_parent_main,
            name_parent_scnd=           name_parent_scnd,
            name_child=                 name_child,
            save_topdir_parent_main=    save_topdir_parent_main,
            save_topdir_parent_scnd=    save_topdir_parent_scnd,
            save_topdir_child=          save_topdir_child,
            save_fn_pfx=                save_fn_pfx,
            logger=                     logger,
            loglevel=                   loglevel)

        if do_gx_ckpt:
            cls.gx_ckpt(
                name_A=             name_parent_main,
                name_B=             name_parent_scnd or name_parent_main,
                name_child=         name_child,
                save_topdir_A=      save_topdir_parent_main,
                save_topdir_B=      save_topdir_parent_scnd,
                save_topdir_child=  save_topdir_child,
                ratio=              ratio,
                noise=              noise)
        else:

            # wrapped into subprocess for better separation of torch objects (similar to mrg_ckpts() in base_elements)
            @proc_wait
            def build_with_subprocess():
                mod = cls(
                    name=               name_child,
                    save_topdir=        save_topdir_child or save_topdir_parent_main,
                    save_fn_pfx=        save_fn_pfx,
                    logger=             logger,
                    loglevel=           loglevel)
                mod.save() # save checkpoint

            build_with_subprocess()

    # ***************************************************************************************************** train / test

    # converts and loads data to Batcher
    def load_data(
            self,
            data_TR: Dict[str,np.ndarray],
            data_VL: Optional[Dict[str,np.ndarray]]=    None,
            data_TS: Optional[Dict[str,np.ndarray]]=    None,
            split_VL: float=                            0.0,
            split_TS: float=                            0.0):

        data_TR = {k: self.convert(data_TR[k]) for k in data_TR}
        if data_VL:
            data_VL = {k: self.convert(data_VL[k]) for k in data_VL}
        if data_TS:
            data_TS = {k: self.convert(data_TS[k]) for k in data_TS}

        self._batcher = Batcher(
            data_TR=        data_TR,
            data_VL=        data_VL,
            data_TS=        data_TS,
            split_VL=       split_VL,
            split_TS=       split_TS,
            batch_size=     self.batch_size,
            batching_type=  'random_cov',
            seed=           self.seed,
            logger=         get_child(self._log, 'Batcher'))

    # trains model, returns optional test score
    def run_train(
            self,
            data_TR: Dict[str,np.ndarray],  # INFO: it will also accept Dict[str,torch.Tensor] :) !
            data_VL: Optional[Dict[str,np.ndarray]]=    None,
            data_TS: Optional[Dict[str,np.ndarray]]=    None,
            split_VL: float=                            0.0,
            split_TS: float=                            0.0,
            n_batches: Optional[int]=                   None,
            test_freq=                                  100,    # number of batches between tests, model SHOULD BE tested while training
            mov_avg_factor=                             0.1,
            save_max=                                   True,   # allows to save model while training (after max test)
            use_F1=                                     True,   # uses F1 as a train/test score (not acc)
        ) -> Optional[float]:

        if data_TR:
            self.load_data(
                data_TR=    data_TR,
                data_VL=    data_VL,
                data_TS=    data_TS,
                split_VL=   split_VL,
                split_TS=   split_TS)

        if not self._batcher: raise MOTorchException(f'{self.name} has not been given data for training, use load_data()')

        self._log.info(f'{self.name} - training starts [acc / F1 / loss]')
        self._log.info(f'data sizes (TR,VL,TS) samples: {self._batcher.get_data_size()}')

        if n_batches is None: n_batches = self.n_batches  # take default
        self._log.info(f'batch size:             {self["batch_size"]}')
        self._log.info(f'train for num_batches:  {n_batches}')

        batch_IX = 0                            # this loop (local) batch counter
        tr_accL = []
        tr_f1L = []
        tr_lssL = []

        score_name = 'F1' if use_F1 else 'acc'
        ts_score_max = 0                        # test score (acc or F1) max
        ts_score_all_results = []               # test score all results
        ts_score_mav = MovAvg(mov_avg_factor)   # test score (acc or F1) moving average

        # initial save
        if not self.read_only and save_max:
            self.save_ckpt()

        ts_bIX = [bIX for bIX in range(n_batches+1) if not bIX % test_freq] # batch indexes when test will be performed
        assert ts_bIX, 'ERR: model SHOULD BE tested while training!'
        ten_factor = int(0.1*len(ts_bIX)) # number of tests for last 10% of training
        if ten_factor < 1: ten_factor = 1 # we need at least one result
        if self.hpmser_mode: ts_bIX = ts_bIX[-ten_factor:]

        while batch_IX < n_batches:

            out = self.backward(**self._batcher.get_batch(), bypass_data_conv=True)

            loss = out['loss']
            acc = out['acc'] if 'acc' in out else None
            f1 = out['f1'] if 'f1' in out else None

            batch_IX += 1

            if self.do_TB:
                self.log_TB(value=loss,                tag='tr/loss',    step=self.train_step)
                self.log_TB(value=out['gg_norm'],      tag='tr/gn',      step=self.train_step)
                self.log_TB(value=out['gg_norm_clip'], tag='tr/gn_clip', step=self.train_step)
                self.log_TB(value=out['currentLR'],    tag='tr/cLR',     step=self.train_step)
                if acc is not None:
                    self.log_TB(value=acc,             tag='tr/acc',     step=self.train_step)
                if f1 is not None:
                    self.log_TB(value=f1,              tag='tr/F1',      step=self.train_step)

            if acc is not None: tr_accL.append(acc)
            if f1 is not None: tr_f1L.append(f1)
            tr_lssL.append(loss)

            if batch_IX in ts_bIX:

                ts_loss, ts_acc, ts_f1 = self.run_test()

                ts_score = ts_f1 if use_F1 else ts_acc
                if ts_score is not None:
                    ts_score_all_results.append(ts_score)
                if self.do_TB:
                    if ts_loss is not None:
                        self.log_TB(value=ts_loss,                    tag='ts/loss',              step=self.train_step)
                    if ts_acc is not None:
                        self.log_TB(value=ts_acc,                     tag='ts/acc',               step=self.train_step)
                    if ts_f1 is not None:
                        self.log_TB(value=ts_f1,                      tag='ts/F1',                step=self.train_step)
                    if ts_score is not None:
                        self.log_TB(value=ts_score_mav.upd(ts_score), tag=f'ts/{score_name}_mav', step=self.train_step)

                tr_acc_nfo = f'{100*sum(tr_accL)/test_freq:.1f}' if acc is not None else '--'
                tr_f1_nfo =  f'{100*sum(tr_f1L)/test_freq:.1f}' if f1 is not None else '--'
                tr_loss_nfo = f'{sum(tr_lssL)/test_freq:.3f}'
                ts_acc_nfo = f'{100*ts_acc:.1f}' if ts_acc is not None else '--'
                ts_f1_nfo = f'{100*ts_f1:.1f}' if ts_f1 is not None else '--'
                ts_loss_nfo = f'{ts_loss:.3f}' if ts_loss is not None else '--'
                self._log.info(f'# {self["train_step"]:5d} TR: {tr_acc_nfo} / {tr_f1_nfo} / {tr_loss_nfo} -- TS: {ts_acc_nfo} / {ts_f1_nfo} / {ts_loss_nfo}')
                tr_accL = []
                tr_f1L = []
                tr_lssL = []

                # model is saved for max ts_score
                if ts_score is not None and ts_score > ts_score_max:
                    ts_score_max = ts_score
                    if not self.read_only and save_max:
                        self.save_ckpt()

        self._log.info(f'### model {self.name} finished training')

        ts_score_fin = None

        if save_max:
            ts_score_fin = ts_score_max
            self.load_ckpt()

        # weighted (linear ascending weight) test score for last 10% test results
        else:
            if ts_score_all_results:
                ts_score_fin = 0.0
                weight = 1
                sum_weight = 0
                for tr in ts_score_all_results[-ten_factor:]:
                    ts_score_fin += tr*weight
                    sum_weight += weight
                    weight += 1
                ts_score_fin /= sum_weight

        if ts_score_fin is not None:
            self._log.info(f' > test_{score_name}_max: {ts_score_max:.4f}')
            self._log.info(f' > test_{score_name}_fin: {ts_score_fin:.4f}')
            if self.do_TB:
                self.log_TB(value=ts_score_fin, tag=f'ts/ts_{score_name}_fin', step=self.train_step)

        return ts_score_fin

    # tests model, returns: optional loss (average), optional accuracy, optional F1
    # optional loss <- since there may be not TS batches
    def run_test(
            self,
            data: Optional[Dict[str,np.ndarray]]=   None,
            split_TS: float=                        1.0, # if data for test will be given above, by default MOTorch will be tested on ALL
    ) -> Tuple[Optional[float], Optional[float], Optional[float]]:

        if data:
            self.load_data(data_TR=data, split_TS=split_TS)

        if not self._batcher: raise MOTorchException(f'{self.name} has not been given data for testing, use load_data() or give it while testing!')

        batches = self._batcher.get_TS_batches()
        lossL = []
        accL = []
        f1L = []
        n_all = 0
        for batch in batches:
            out = self.loss(**batch, bypass_data_conv=True)
            n_new = len(out['logits'])
            n_all += n_new
            lossL.append(out['loss']*n_new)
            if 'acc' in out: accL.append(out['acc']*n_new)
            if 'f1' in out:  f1L.append(out['f1']*n_new)

        acc_avg = sum(accL)/n_all if accL else None
        f1_avg = sum(f1L)/n_all if f1L else None
        loss_avg = sum(lossL)/n_all if lossL else None
        return loss_avg, acc_avg, f1_avg


    # *********************************************************************************************** other / properties

    # updates scheduler baseLR of 0 group
    def update_baseLR(self, lr: float):
        self.baseLR = lr # in case model will be saved >> loaded
        self._scheduler.update_base_lr0(lr)

    @property
    def module(self):
        return self._module

    @property
    def tbwr(self):
        return self._TBwr

    # logs value to TB
    def log_TB(
            self,
            value,
            tag: str,
            step: int) -> None:
        if self.do_TB: self._TBwr.add(value=value, tag=tag, step=step)
        else: self._log.warning(f'{self.name} cannot log to TensorBoard since \'do_TB\' flag was set to False!')

    @property
    def logger(self):
        return self._log

    @property
    def size(self) -> int:
        pp = 0
        for p in list(self.parameters()):
            nn = 1
            for s in list(p.size()):
                nn = nn * s
            pp += nn
        return pp

    def __str__(self):
        s = f'MOTorch: {ParaSave.__str__(self)}\n'
        s += str(self._module)
        return s


import torch
import unittest

from torchness.base_elements import my_initializer


class TestBaseElements(unittest.TestCase):

    def test_my_initializer(self):
        tns = torch.zeros(1000)
        #print(tns)
        my_initializer(tns, std=0.1)
        #print(tns)
        print(tns.numpy().std())
        self.assertTrue(0.08 < tns.numpy().std() < 0.12)


import numpy as np
from pypaq.mpython.mpdecor import proc_wait
from pypaq.lipytools.pylogger import get_pylogger
import torch
import unittest

from torchness.motorch import MOTorch, Module, MOTorchException
from torchness.layers import LayDense

from tests.envy import flush_tmp_dir

MOTORCH_DIR = f'{flush_tmp_dir()}/motorch'
MOTorch.SAVE_TOPDIR = MOTORCH_DIR


class LinModel(Module):

    def __init__(
            self,
            in_drop: float,
            in_shape=   784,
            out_shape=  10,
            loss_func=  torch.nn.functional.cross_entropy,
            device=     None,
            dtype=      None,
            seed=       121,
            **kwargs,
    ):

        Module.__init__(self, **kwargs)

        self.in_drop_lay = torch.nn.Dropout(p=in_drop) if in_drop>0 else None
        self.lin = LayDense(in_features=in_shape, out_features=out_shape)
        self.loss_func = loss_func

        self.logger.debug('LinModel initialized!')

    def forward(self, inp) -> dict:
        if self.in_drop_lay is not None: inp = self.in_drop_lay(inp)
        logits = self.lin(inp)
        return {'logits': logits}

    def loss(self, inp, lbl) -> dict:
        out = self(inp)
        out['loss'] = self.loss_func(out['logits'], lbl)
        out['acc'] = self.accuracy(out['logits'], lbl)  # using baseline
        return out


logger = get_pylogger(name='test_motorch', level=20)


class TestMOTorch(unittest.TestCase):

    def setUp(self) -> None:
        flush_tmp_dir()


    def test_base_init(self):
        logger.setLevel(10)
        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.0,
            logger=         logger)
        print(model)
        logger.setLevel(20)


    def test_init_raises(self):
        self.assertRaises(Exception, MOTorch)
        kwargs = dict(name='LinModel')
        self.assertRaises(Exception, MOTorch, **kwargs)
        kwargs = dict(module_type=LinModel)
        self.assertRaises(Exception, MOTorch, **kwargs)


    def test_device(self):

        # force GPU
        model = MOTorch(
            module_type=    LinModel,
            device=         0,
            in_drop=        0.0,
            logger=         logger)
        dev = model.device
        print(dev)
        self.assertTrue(dev == 'cuda:0' if torch.cuda.is_available() else 'cpu')

        # GPU or CPU
        model = MOTorch(
            module_type=    LinModel,
            device=         -1,
            in_drop=        0.0,
            logger=         logger)
        dev = model.device
        print(dev)
        self.assertTrue('cuda' in dev or 'cpu' in dev)


    def test_logger(self):
        log = get_pylogger(
            name=       'test_motorch',
            level=      10,
            flat_child= True)
        model = MOTorch(
            module_type=    LinModel,
            device=         0,
            in_drop=        0.0,
            logger=         log)
        model.save()


    def test_save_load(self):
        model = MOTorch(
            module_type=    LinModel,
            loglevel=       10,
            in_drop=        0.1,
            logger=         logger)
        self.assertTrue(model['seed']==121 and model['baseLR']==0.0003)
        self.assertTrue('loss' not in model.get_managed_params())
        model.save()
        name = model.name

        print('\nsaved, now loading..')

        logger.setLevel(10)
        model = MOTorch(
            name=       name,
            loglevel=   10,
            logger=     logger)
        self.assertTrue(model['in_drop']==0.1 and model['in_shape']==784)
        logger.setLevel(10)


    def test_base_creation_and_call(self):

        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.1,
            logger=         logger)

        inp = np.random.random((5,784)).astype(np.float32)
        lbl = np.random.randint(0,9,5)

        out = model(inp)
        print(out)
        logits = out['logits']
        self.assertTrue(logits.shape[0]==5 and logits.shape[1]==10)

        out = model.loss(inp, lbl)
        loss = out['loss']
        acc = out['acc']
        print(loss, acc)
        self.assertTrue(type(loss) is torch.Tensor)
        self.assertTrue(type(acc) is float)

        for _ in range(5):
            out = model.backward(inp, lbl)
            loss = out['loss']
            acc = out['acc']
            print(model.train_step, loss, acc)


    def test_saved_step(self):

        model = MOTorch(
            name=           'modA',
            module_type=    LinModel,
            in_drop=        0.1,
            logger=         logger)

        inp = np.random.random((5, 784)).astype(np.float32)
        lbl = np.random.randint(0, 9, 5)
        for _ in range(5):
            out = model.backward(inp, lbl)
        print(model.name, model.train_step)
        model.save()

        model = MOTorch(name=model.name, logger=logger)
        print(model.name, model.train_step)
        self.assertTrue(model.name == 'modA' and model.train_step == 5)


    def test_class_method(self):

        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.1,
            logger=         logger)
        print(model.name)
        point_org = model.get_point()
        print(point_org)
        self.assertTrue(point_org['gc_first_avg'])
        model.save()

        point = MOTorch.load_point(name=model.name)
        print(point)
        self.assertTrue(not point['gc_first_avg']) # INFO: gc_first_avg is updated while MOTorch.save()

        point_org.pop('gc_first_avg')
        point.pop('gc_first_avg')
        self.assertTrue(point_org == point)


    def test_inherited_interfaces(self):

        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.1,
            logger=         logger)
        print(model.parameters())
        model.float()
        print(model.state_dict())
        print(model.get_point())


    def test_training_mode(self):

        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.8,
            logger=         logger)
        print(model.training)

        inp = np.random.random((5,784)).astype(np.float32)
        lbl = np.random.randint(0,9,5)

        out = model(inp)
        print(out)
        out = model.loss(inp, lbl)
        loss = out['loss']
        print(loss)

        out = model(inp, set_training=True)
        print(out)
        out = model.loss(inp, lbl, set_training=True)
        loss = out['loss']
        print(loss)


    def test_name_stamp(self):

        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.1,
            logger=         logger)
        print(model['name'])
        self.assertTrue(model['name'] == 'LinModel_MOTorch')

        model = MOTorch(
            module_type=    LinModel,
            name=           'LinTest',
            in_drop=        0.1,
            logger=         logger)
        print(model['name'])
        self.assertTrue(model['name'] == 'LinTest')

        model = MOTorch(
            module_type=        LinModel,
            name_timestamp=     True,
            in_drop=            0.1,
            logger=             logger)
        print(model['name'])
        self.assertTrue(model['name'] != 'MOTorch_LinModel')
        self.assertTrue({d for d in '0123456789'} & set([l for l in model['name']]))


    def test_ParaSave(self):

        model = MOTorch(
            module_type=    LinModel,
            loglevel=       20,
            in_shape=       12,
            out_shape=      12,
            in_drop=        0.0,
            logger=         logger)
        pms = model.get_managed_params()
        print(f'model.get_managed_params(): {pms}')
        orig_seed = model['seed']
        print(f'orig_seed: {orig_seed}')
        model.save()

        MOTorch.oversave_point(
            name=   model["name"],
            seed=   252)

        # this will not load
        dna = model.load_point(
            name=           model["name"],
            save_topdir=    'other')
        print(dna)
        self.assertFalse(dna)

        # this will load
        dna = model.load_point(name=model["name"])
        print(dna)
        for p in pms:
            if p not in dna: print(p)
            self.assertTrue(p in dna)
        self.assertTrue(dna["seed"]==252)

        # this model will not load
        model = MOTorch(
            name=           'inne',
            module_type=    LinModel,
            in_drop=        0.0,
            logger=         logger)
        print(f'not loaded model in_shape: {model["in_shape"]}')
        self.assertTrue(model['in_shape'] != 12)

        # this model will load from MOTORCH_DIR
        model = MOTorch(
            module_type=    LinModel,
            loglevel=       10,
            logger=         logger)
        print(model['in_shape'])
        self.assertTrue(model['in_shape'] == 12)

        model = MOTorch(
            module_type=    LinModel,
            name_timestamp= True,
            in_shape=       12,
            out_shape=      12,
            in_drop=        0.0,
            logger=         logger)
        model.save()
        print(model['name'])

        model = MOTorch(name=model['name'], logger=logger)
        self.assertTrue(model['in_shape'] == 12)

        model.copy_saved_point(
            name_src=           model["name"],
            name_trg=           'CopiedPS')

        model.gx_saved_point(
            name_parent_main=           model["name"],
            name_parent_scnd=           None,
            name_child=                 'GXed')

        psdd = {'seed': [0,1000]}
        model = MOTorch(
            module_type=    LinModel,
            name=           'GXLin',
            psdd=           psdd,
            in_drop=        0.0,
            logger=         logger)

        print(model.gxable)
        print(model['psdd'])
        print(model['seed'])
        dna = model.gx_point(
            parent_main=    model,
            prob_noise=     0.0,
            prob_axis=      0.0)
        print(dna['seed'])
        dna = model.gx_point(
            parent_main=    model,
            prob_noise=     1.0,
            prob_axis=      1.0)
        print(dna['seed'])


    def test_params_resolution(self):

        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.1,
            logger=         logger)
        print(model['seed'])        # value from MOTORCH_DEFAULTS
        print(model['in_shape'])    # value from module_type defaults
        self.assertTrue(model['seed'] == 121)
        self.assertTrue(model['in_shape'] == 784)

        model = MOTorch(
            module_type=    LinModel,
            seed=           151,
            in_drop=        0.1,
            logger=         logger)
        print(model['seed'])        # MOTORCH_DEFAULTS overridden with kwargs
        self.assertTrue(model['seed'] == 151)

        model = MOTorch(
            module_type=    LinModel,
            in_shape=       24,
            out_shape=      24,
            in_drop=        0.1,
            logger=         logger)
        print(model['seed'])        # MOTORCH_DEFAULTS overridden with module_type defaults
        self.assertTrue(model['seed'] == 121)
        model.save()
        name = model.name

        model = MOTorch(
            name=   name,
            seed=   212,
            logger= logger)
        print(model['in_shape'])    # loaded from save
        print(model['seed'])        # saved overridden with kwargs
        self.assertTrue(model['in_shape'] == 24)
        self.assertTrue(model['seed'] == 212)


    def test_seed_of_torch(self):

        model = MOTorch(
            module_type=    LinModel,
            seed=       121,
            in_drop=    0.1,
            logger=     logger)

        inp = np.random.random((5,784)).astype(np.float32)
        out1 = model(inp)
        print(model['seed'])
        print(out1)

        model = MOTorch(
            module_type=    LinModel,
            seed=           121,
            in_drop=        0.1,
            logger=         logger)

        out2 = model(inp)
        print(model['seed'])
        print(out2)

        self.assertTrue(np.sum(out1['logits'].cpu().detach().numpy()) == np.sum(out2['logits'].cpu().detach().numpy()))


    def test_read_only(self):

        model = MOTorch(
            module_type=    LinModel,
            in_drop=        0.1,
            logger=         logger)
        model.save()
        name = model.name

        model = MOTorch(
            name=       name,
            read_only=  True,
            logger=     logger)
        self.assertRaises(MOTorchException, model.save)


    def test_save_load_full(self):

        model = MOTorch(
            module_type=    LinModel,
            in_shape=       256,
            out_shape=      10,
            name_timestamp= True,
            seed=           121,
            in_drop=        0.1,
            logger=         logger)
        name = model.name
        print(model.name)

        inp = np.random.random((5, 256)).astype(np.float32)

        out1 = model(inp)
        print(out1)
        model.save()

        loaded_model = MOTorch(
            name=           name,
            seed=           123, # although different seed, model will load checkpoint
            logger=         logger)
        print(loaded_model.name)
        out2 = loaded_model(inp)
        print(out2)
        # print(loaded_model)

        self.assertTrue(np.sum(out1['logits'].cpu().detach().numpy()) == np.sum(out2['logits'].cpu().detach().numpy()))


    def test_copy_saved(self):

        model = MOTorch(
            module_type=        LinModel,
            in_shape=       256,
            out_shape=      10,
            name_timestamp= True,
            seed=           121,
            in_drop=        0.1,
            logger=         logger)
        name = model.name
        print(model)
        model.save()

        name_copied = f'{name}_copied'
        MOTorch.copy_saved(
            name_src=           name,
            name_trg=           name_copied)

        model = MOTorch(name=name_copied, logger=logger)
        print(model)


    def test_hpmser_mode(self):

        model = MOTorch(
            module_type=    LinModel,
            hpmser_mode=    True,
            in_drop=        0.1,
            logger=         logger)
        self.assertRaises(MOTorchException, model.save)

# those two needs to be run separately, idnk why..
class TestMOTorch_GX(unittest.TestCase):

    def setUp(self) -> None:
        flush_tmp_dir()


    def test_gx_ckpt(self):

        name_A = 'modA'
        name_B = 'modB'

        # needs to create in separate process
        @proc_wait
        def create():
            model = MOTorch(
                module_type=    LinModel,
                name=           name_A,
                seed=           121,
                in_drop=        0.1,
                logger=         logger)
            model.save()

            model = MOTorch(
                module_type=    LinModel,
                name=           name_B,
                seed=           121,
                in_drop=        0.1,
                logger=         logger)
            model.save()

        create()

        MOTorch.gx_ckpt(
            name_A=         name_A,
            name_B=         name_B,
            name_child=     f'{name_A}_GXed')


    def test_gx_saved(self):

        name_A = 'modA'
        name_B = 'modB'

        # needs to create in separate process
        @proc_wait
        def create():
            model = MOTorch(
                module_type=    LinModel,
                name=           name_A,
                seed=           121,
                in_drop=        0.1,
                logger=         logger)
            model.save()

            model = MOTorch(
                module_type=    LinModel,
                name=           name_B,
                seed=           121,
                in_drop=        0.1,
                logger=         logger)
            model.save()

        create()

        MOTorch.gx_saved(
            name_parent_main=   name_A,
            name_parent_scnd=   name_B,
            name_child=         f'{name_A}_GXed')



import torch
import unittest

from torchness.base_elements import TorchnessException
from torchness.encoders import LayBlockDRT, EncDRT, LayBlockCNN, EncCNN, LayBlockTNS, EncTNS


class TestEncoders(unittest.TestCase):

    def test_LayBlockDRT_base(self):

        in_width = 10
        inp = torch.rand(in_width) - 0.5
        print(inp, inp.dtype)

        lay_drt = LayBlockDRT(in_width)
        print(lay_drt)
        out = lay_drt(inp)
        print(out)
        self.assertTrue(out['out'].shape[-1] == in_width)
        self.assertTrue(out['zsL'][0].shape[-1] == in_width)


    def test_LayBlockDRT_kwargs(self):

        in_width = 10
        inp = torch.rand(in_width) - 0.5

        lay_drt = LayBlockDRT(
            in_width=       in_width,
            do_scaled_dns=  True,
            dns_scale=      4,
            lay_dropout=    0.1,
            res_dropout=    0.1)
        print(lay_drt)
        out = lay_drt(inp)
        print(out)
        self.assertTrue(out['out'].shape[-1] == in_width)
        self.assertTrue(out['zsL'][0].shape[-1] == in_width * 4)


    def test_LayBlockDRT_device(self):

        in_width = 10
        inp = torch.rand(in_width) - 0.5

        lay_drt = LayBlockDRT(in_width)

        dev = torch.device('cuda')
        lay_drt = lay_drt.to(dev)
        inp = inp.to(dev)
        out = lay_drt(inp)
        print(out['out'].device)
        print(out)

        lay_drt = LayBlockDRT(in_width)
        self.assertRaises(RuntimeError, lay_drt, inp)  # devices mismatch


    def test_LayBlockDRT_double(self):

        in_width = 10
        inp = torch.rand(in_width) - 0.5

        lay_drt = LayBlockDRT(
            in_width=       in_width,
            do_scaled_dns=  True,
            lay_dropout=    0.1,
            res_dropout=    0.1,
            dtype=          torch.double)
        print(lay_drt)
        inp = inp.to(torch.double)
        out = lay_drt(inp)
        print(out['out'].dtype)
        print(out)


    def test_EncDRT(self):

        in_width = 10
        inp = torch.rand(in_width) - 0.5
        print(inp, inp.dtype)

        enc_drt = EncDRT(in_width, shared_lays=True, dns_scale=4)
        print(enc_drt)
        self.assertTrue(len(list(enc_drt.children())) == 2) # ln_in, shared LayDRT
        out = enc_drt(inp)
        print(out)
        self.assertTrue(out['zsL'][0].shape[-1] == in_width*4)

        enc_drt = EncDRT(in_width, lay_width=2*in_width, dns_scale=3)
        print(enc_drt)
        self.assertTrue(len(list(enc_drt.children())) == 8) # projection, ln_in, 6 * LayDRT
        out = enc_drt(inp)
        print(out)
        self.assertTrue(out['zsL'][0].shape[-1] == in_width*3*2)

        enc_drt = EncDRT(
            in_width=       in_width,
            in_dropout=     0.1,
            n_layers=       4,
            lay_width=      16,
            do_scaled_dns=  True,
            dns_scale=      3,
            lay_dropout=    0.3,
            res_dropout=    0.3,
            dtype=          torch.double)
        print(enc_drt)
        self.assertTrue(len(list(enc_drt.children())) == 7) # in_drop, projection, ln_in, 4 * LayDRT
        inp = inp.to(torch.double)
        out = enc_drt(inp)
        print(out)
        self.assertTrue(len(out['zsL']) == 4) # 4 layers
        self.assertTrue(out['zsL'][0].shape[-1] == 16*3) # lay_width * dns_scale

        enc_drt.to(torch.float)
        self.assertRaises(RuntimeError, enc_drt, inp) # expected scalar type Double but found Float


    def test_LayBlockCNN_base_encoder(self):

        n_time = 6
        n_filters = 4
        inp = torch.rand(n_time, n_filters)
        print(f'inp shape {inp.shape}')

        for kernel_size in [3,5,7,9]:
            lay_cnn = LayBlockCNN(n_filters, kernel_size=kernel_size)
            print(lay_cnn)

            out = lay_cnn(inp)
            print(f'out shape {out["out"].shape}')
            print(f'state: {out["state"]}')
            self.assertTrue(inp.shape == out['out'].shape)
            self.assertTrue(out['state'] is None)


    def test_LayBlockCNN_base_casual(self):

        n_time = 6
        n_filters = 4
        inp = torch.rand(n_time, n_filters)
        print(f'inp shape {inp.shape}')

        for kernel_size in [3,5,7,9]:
            lay_cnn = LayBlockCNN(n_filters, kernel_size=kernel_size)
            print(lay_cnn)

            zero_history_def = lay_cnn.get_zero_history()
            zero_history = lay_cnn.get_zero_history(inp)
            self.assertTrue(zero_history_def.shape == zero_history.shape)
            print(f'zero_history shape {zero_history.shape}')

            out = lay_cnn(inp, history=zero_history)
            print(f'out shape {out["out"].shape}')
            print(f'state shape: {out["state"].shape}')
            self.assertTrue(out['out'].shape == inp.shape)
            self.assertTrue(out['state'].shape == zero_history.shape)


    def test_LayBlockCNN_kwargs(self):

        n_filters = 4
        inp = torch.rand(6, n_filters)
        print(inp)

        lay_cnn = LayBlockCNN(
            n_filters=      n_filters,
            lay_dropout=    0.1,
            res_dropout=    0.1,
            do_ldrt=        True)
        print(lay_cnn)
        out = lay_cnn(inp)
        print(out)


    def test_LayBlockCNN_more(self):

        n_time = 32
        n_filters = 64
        inp = torch.rand(16, n_time, n_filters)
        print(f'inp shape {inp.shape}')

        kernel_size = 3
        lay_cnn = LayBlockCNN(n_filters, kernel_size=kernel_size)
        print(lay_cnn)

        zero_history_def = lay_cnn.get_zero_history()
        zero_history = lay_cnn.get_zero_history(inp)
        print(f'zero_history_def shape {zero_history_def.shape}')
        print(f'zero_history shape {zero_history.shape}')

        out = lay_cnn(inp, history=zero_history)
        print(f'out shape {out["out"].shape}')
        print(f'state shape: {out["state"].shape}')
        self.assertTrue(out['out'].shape == inp.shape)
        self.assertTrue(out['state'].shape == zero_history.shape)


    def test_EncCNN_base(self):

        self.assertRaises(TorchnessException, EncCNN, in_features=6, kernel_size=4) # even number for kernel

        in_features = 128
        inp = torch.rand(256,512,in_features)
        print(inp.shape)

        enc = EncCNN(in_features)
        print(enc)
        enc_out = enc(inp)

        print(enc_out['out'].shape)
        self.assertTrue(enc_out['out'].shape == inp.shape)

        self.assertTrue(enc_out['state'] is None)


    def test_EncCNN_casual(self):

        in_features = 128
        inp = torch.rand(256,512,in_features)
        print(inp.shape)

        enc = EncCNN(in_features)
        print(enc)

        zero_history_base = enc.get_zero_history()
        print(zero_history_base.shape)
        zero_history = enc.get_zero_history(inp)
        enc_out = enc(inp, history=zero_history)

        print(enc_out['out'].shape)
        self.assertTrue(enc_out['out'].shape == inp.shape)

        print(enc_out['state'].shape)
        self.assertTrue(enc_out['state'].shape == zero_history.shape)


    def test_EncCNN_kwargs(self):

        n_filters = 48
        in_features = 32
        inp = torch.rand(18,96,in_features)
        print(inp.shape)
        inp = inp.to(torch.double)

        enc = EncCNN(
            in_features=        in_features,
            time_drop=          0.1,
            feat_drop=          0.2,
            n_layers=           5,
            kernel_size=        7,
            n_filters=          n_filters,
            lay_dropout=        0.15,
            res_dropout=        0.25,
            do_ldrt=            True,
            ldrt_drop=          0.05,
            ldrt_res_dropout=   0.07,
            dtype=              torch.double)
        print(enc)
        enc_out = enc(inp)

        print(enc_out['out'].shape)
        in_sh = list(inp.shape)
        in_sh[-1] = n_filters
        self.assertTrue(list(enc_out['out'].shape) == in_sh)

        zero_history = enc.get_zero_history(inp)
        print(zero_history.shape)
        self.assertTrue(list(zero_history.shape) == [18,5,6,48])


    def test_EncCNN_shared(self):

        self.assertRaises(TorchnessException, EncCNN, in_features=6, kernel_size=4) # even number for kernel

        in_features = 128
        inp = torch.rand(256,512,in_features)
        print(inp.shape)

        enc = EncCNN(in_features, shared_lays=True)
        print(enc)
        enc_out = enc(inp)

        print(enc_out['out'].shape)
        self.assertTrue(list(enc_out['out'].shape) == list(inp.shape))

        zero_history = enc.get_zero_history(inp)
        print(zero_history.shape)
        self.assertTrue(list(zero_history.shape) == [256,6,2,128])


    def test_LayBlockTNS_base(self):

        in_features = 64
        inp = torch.rand(16,32,in_features)
        print(inp.shape)

        lay_tns = LayBlockTNS(d_model=in_features)
        print(lay_tns)
        out = lay_tns(inp)
        print(out['out'].shape)
        print(out['zsL'][0].shape)

        # TAT
        query = torch.mean(inp, dim=-2, keepdim=True)
        print(query.shape)
        out = lay_tns(inp, task_query=query)
        print(out['out'].shape)
        print(out['zsL'][0].shape)


    def test_EncTNS_base(self):

        in_features = 64
        inp = torch.rand(16,32,in_features)
        print(inp.shape)

        enc = EncTNS(num_layers=4, d_model=in_features)
        self.assertTrue(enc.pos_emb is None)
        print(enc)
        out = enc(inp)
        print(out['out'].shape, len(out['zsL']))


    def test_EncTNS_PE(self):

        in_features = 64
        seq_len = 32
        inp = torch.rand(16,seq_len,in_features)
        print(inp.shape)

        enc = EncTNS(num_layers=2, d_model=in_features, max_seq_len=48)
        print(enc)
        print(enc.pos_emb)
        self.assertTrue(enc.pos_emb is not None)
        out = enc(inp)
        print(out['out'].shape)


    def test_EncRNS_TAT(self):

        in_features = 64
        seq_len = 32
        inp = torch.rand(16, seq_len, in_features)
        print(inp.shape)

        enc = EncTNS(
            num_layers=     4,
            num_layers_TAT= 2,
            d_model=        in_features)
        print(enc)
        out = enc(inp)
        print(out['out'].shape)


    def test_EncRNS_TAT_shared(self):

        in_features = 64
        seq_len = 32
        inp = torch.rand(16, seq_len, in_features)
        print(inp.shape)

        enc = EncTNS(
            num_layers=     4,
            num_layers_TAT= 2,
            shared_lays=    (3,1,2),
            d_model=        in_features)
        print(enc)
        out = enc(inp)
        print(out['out'].shape)



import torch
from typing import Optional, Tuple, Union

from torchness.types import ACT, INI, TNS, DTNS
from torchness.base_elements import bert_initializer, my_initializer, TorchnessException
from torchness.layers import LayDense, TF_Dropout, LayConv1D, LayRES, zeroes


# Block (Layer) of EncDRT (LN > Dense or two_with_drop_in_between > drop > RES with drop)
class LayBlockDRT(torch.nn.Module):

    def __init__(
            self,
            in_width: int,
            do_scaled_dns: bool=        False,          # two denses (True) or single dense (False)
            dns_scale: int=             4,              # up-scale for first dense of two
            activation: ACT=            torch.nn.ReLU,
            interlay_dropout: float=    0.0,            # dropout in between two denses
            lay_dropout: float=         0.0,            # dropout after dense/s
            residual: bool=             True,           # residual yes/no
            res_dropout: float=         0.0,            # dropout on residual connection
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(LayBlockDRT, self).__init__()

        if initializer is None: initializer = my_initializer

        self.ln_in = torch.nn.LayerNorm(
            normalized_shape=   in_width,
            device=             device,
            dtype=              dtype)

        self.denses = []
        if do_scaled_dns:
            # dense (scale up) with activation
            self.denses.append(LayDense(
                in_features=    in_width,
                out_features=   in_width * dns_scale,
                activation=     activation,
                bias=           True,
                device=         device,
                dtype=          dtype,
                initializer=    initializer))
            # dense (scale down) without activation
            self.denses.append(LayDense(
                in_features=    in_width * dns_scale,
                out_features=   in_width,
                activation=     None,
                bias=           True,
                device=         device,
                dtype=          dtype,
                initializer=    initializer))
        else:
            # just single dense, with activation
            self.denses.append(LayDense(
                in_features=    in_width,
                out_features=   in_width,
                activation=     activation,
                bias=           True,
                device=         device,
                dtype=          dtype,
                initializer=    initializer))
        for dix, l in enumerate(self.denses): self.add_module(f'dense{dix}', l)

        self.drop_interlay = torch.nn.Dropout(p=interlay_dropout) if interlay_dropout and do_scaled_dns else None

        self.drop_lay = torch.nn.Dropout(p=lay_dropout) if lay_dropout else None

        self.res = LayRES(in_features=in_width, dropout=res_dropout) if residual else None

    def forward(self, inp:TNS) -> DTNS:

        out = self.ln_in(inp)

        out = self.denses[0](out)
        zsL = [zeroes(out)]

        if len(self.denses) > 1: # there is second one, without activation

            if self.drop_interlay:
                out = self.drop_interlay(out)

            out = self.denses[1](out)

        if self.drop_lay:
            out = self.drop_lay(out)

        if self.res:
            out = self.res(inp=out, bypass=inp)

        return {
            'out':  out,
            'zsL':  zsL}


# Deep Residual encoder based on stacked LayBlockDRT
class EncDRT(torch.nn.Module):

    def __init__(
            self,
            in_width: int,
            in_dropout: float=          0.0,            # dropout on input
            n_layers: int=              6,
            shared_lays: bool=          False,          # shared variables in enc_layers
            lay_width: Optional[int]=   None,           # for None matches input width
            do_scaled_dns: bool=        True,
            dns_scale: int=             4,              # scale(*) of first dense
            activation: ACT=            torch.nn.ReLU,  # gelu is really worth a try
            lay_dropout: float=         0.0,            # dropout after two denses
            residual: bool=             True,           # residual yes/no
            res_dropout: float=         0.0,            # dropout on residual connection
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(EncDRT, self).__init__()

        if initializer is None: initializer = my_initializer

        self.in_drop_lay = torch.nn.Dropout(p=in_dropout) if in_dropout else None

        self.in_width = in_width
        self.lay_width = lay_width or self.in_width
        self.projection_lay = LayDense(
            in_features=    self.in_width,
            out_features=   self.lay_width,
            activation=     None,
            bias=           False,
            device=         device,
            dtype=          dtype,
            initializer=    initializer) if self.lay_width != self.in_width else None

        self.ln_in = torch.nn.LayerNorm(
            normalized_shape=   self.lay_width,
            device=             device,
            dtype=              dtype)

        num_layers_to_build = 1 if shared_lays else n_layers
        self.drt_lays = [LayBlockDRT(
            in_width=       self.lay_width,
            do_scaled_dns=  do_scaled_dns,
            dns_scale=      dns_scale,
            activation=     activation,
            lay_dropout=    lay_dropout,
            residual=       residual,
            res_dropout=    res_dropout,
            device=         device,
            dtype=          dtype,
            initializer=    initializer
        ) for _ in range(num_layers_to_build)]
        for lix,lay in enumerate(self.drt_lays): self.add_module(f'lay_drt_{lix}',lay)
        if shared_lays and n_layers > 1: self.drt_lays *= n_layers

    def forward(self, inp:TNS) -> DTNS:

        zsL = []

        out = inp

        if self.in_drop_lay: # input dropout
            out = self.in_drop_lay(out)

        if self.projection_lay: # input projection, no activation <- do not catch zeroes
            out = self.projection_lay(out)

        out = self.ln_in(out)

        for drt_lay in self.drt_lays:
            lay_out = drt_lay(out)
            out = lay_out['out']
            zsL += lay_out['zsL']

        return {
            'out':  out,
            'zsL':  zsL}


# Block (Layer) of EncCNN (LN > CNN > act > drop > RES > LayBlockDRT), number of parameters: kernel*in_features*n_filters
class LayBlockCNN(torch.nn.Module):

    def __init__(
            self,
            n_filters: int,                             # num of filters
            padded=                     True,           # if not padded reduces sequence length
            kernel_size: int=           3,              # layer kernel
            activation: ACT=            torch.nn.ReLU,  # global enc activation func
            lay_dropout: float=         0.0,
            res_dropout: float=         0.0,
            # lay_DRT
            do_ldrt=                    False,          # lay DRT - build or not
            ldrt_do_scaled_dns: bool=   True,
            ldrt_dns_scale: int=        4,
            ldrt_drop: float or None=   0.0,
            ldrt_residual: bool=        True,
            ldrt_res_dropout: float=    0.0,
            # other
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(LayBlockCNN, self).__init__()

        self.n_filters = n_filters
        self.padded = padded
        if kernel_size % 2 == 0: raise TorchnessException('LayBlockCNN kernel_size cannot be even number')
        self.kernel_size = kernel_size

        self.device = device
        self.dtype = dtype

        self.lay_ln = torch.nn.LayerNorm(
            normalized_shape=   self.n_filters,
            device=             self.device,
            dtype=              self.dtype)

        self.lay_conv1D = LayConv1D(
            in_features=    self.n_filters,
            n_filters=      self.n_filters,
            kernel_size=    self.kernel_size,
            padding=        'valid',
            device=         self.device,
            dtype=          self.dtype,
            activation=     None,
            initializer=    initializer)

        self.activation = activation() if activation else None

        self.lay_drop = torch.nn.Dropout(p=lay_dropout) if lay_dropout else None

        self.res = LayRES(in_features=self.n_filters, dropout=res_dropout)

        self.lay_DRT = LayBlockDRT(
            in_width=       self.n_filters,
            do_scaled_dns=  ldrt_do_scaled_dns,
            dns_scale=      ldrt_dns_scale,
            activation=     activation,
            lay_dropout=    ldrt_drop,
            residual=       ldrt_residual,
            res_dropout=    ldrt_res_dropout,
            device=         self.device,
            dtype=          self.dtype,
            initializer=    initializer) if do_ldrt else None

    # prepares baseline 2-dim zero_history
    def _get_zero_history_base(self) -> TNS:
        in_sh = [self.kernel_size-1, self.n_filters]
        return torch.zeros(in_sh).to(self.device, self.dtype)

    # prepares initial history for casual mode, history has shape [.., kernel_size-1, n_filters]
    def get_zero_history(self, inp:Optional[TNS]=None) -> TNS:
        zhb = self._get_zero_history_base()
        if inp is None: return zhb
        else:           return zhb.expand(list(inp.shape)[:-2] + list(zhb.shape))


    def forward(
            self,
            inp: TNS, # INFO: at least 2-dim tensor: [.., seq, feats]
            history: Optional[TNS]= None # INFO: history must be given (even zero_history) for casual mode
    ) -> DTNS:

        zsL = []
        out = self.lay_ln(inp)

        ### pad out and prepare state for return

        state = None
        if self.padded:

            proper_history_shape = list(inp.shape)[:-2] + [self.kernel_size-1, self.n_filters]

            # pad with zeroes on both sides (encoder), ..but no state to return
            if history is None:
                pad_shape = proper_history_shape
                pad_shape[-2] = pad_shape[-2] // 2
                pad = torch.zeros(pad_shape).to(self.device, self.dtype)
                conc = [pad, out, pad]
                out = torch.concat(conc, dim=-2)

            # concatenate with history on the left (casual block)
            else:

                if not list(history.shape) == proper_history_shape:
                    raise TorchnessException(f'wrong histy shape, given {history.shape}, should be {proper_history_shape}')
                conc = [history, out]
                out = torch.concat(conc, dim=-2)

                # prepare state
                inp_for_state = inp
                if inp_for_state.size(-2) < proper_history_shape[-2]: # inp is shorter than history
                    pad_shape = list(inp_for_state.shape)
                    pad_shape[-2] = proper_history_shape[-2] - inp_for_state.size(-2)
                    pad = torch.zeros(pad_shape).to(inp)
                    inp_for_state = torch.concat([pad,inp_for_state], dim=-2)
                sections = [inp_for_state.size(-2)-proper_history_shape[-2], proper_history_shape[-2]]
                state_spl = torch.split(
                    tensor=                 inp_for_state,
                    split_size_or_sections= sections,
                    dim=                    -2)
                state = state_spl[-1]

        out = self.lay_conv1D(out)

        if self.activation:
            out = self.activation(out)
            zsL.append(zeroes(out))

        if self.lay_drop:
            out = self.lay_drop(out)

        # it is not possible to do RES for not padded version
        if self.padded:
            out = self.res(inp=out, bypass=inp)

        if self.lay_DRT:
            lay_out = self.lay_DRT(out)
            out = lay_out['out']
            zsL += lay_out['zsL']

        return {
            'out':      out,
            'state':    state,
            'zsL':      zsL}


# CNN 1D Encoder (for sequences), number of parameters: projection + n_layers*LayBlockCNN
class EncCNN(torch.nn.Module):

    def __init__(
            self,
            in_features: int,                           # input num of channels
            time_drop: float=           0.0,
            feat_drop: float=           0.0,
            # layer
            shared_lays: bool=          False,          # shared variables between LayBlockCNN
            n_layers :int=              6,              # num of layers
            padded=                     True,           # if not padded reduces sequence length
            kernel_size :int=           3,              # layer kernel
            n_filters :Optional[int]=   None,           # num of filters, for None uses in_features
            activation: ACT=            torch.nn.ReLU,  # global enc activation func
            lay_dropout: float=         0.0,
            res_dropout: float=         0.0,
            # lay_DRT
            do_ldrt=                    False,          # lay DRT - build or not
            ldrt_do_scaled_dns: bool=   True,
            ldrt_dns_scale: int=        4,
            ldrt_drop: float or None=   0.0,
            ldrt_residual: bool=        True,
            ldrt_res_dropout: float=    0.0,
            # other
            device=                     None,
            dtype=                      None,
            initializer: INI=           None):

        super(EncCNN, self).__init__()

        self.in_features = in_features
        self.n_layers = n_layers
        self.kernel_size = kernel_size
        self.n_filters = n_filters or self.in_features

        self.device = device
        self.dtype = dtype

        self.in_TFdrop_lay = TF_Dropout(
            time_drop=  time_drop,
            feat_drop=  feat_drop) if time_drop or feat_drop else None

        self.projection_lay = LayDense(
            in_features=    self.in_features,
            out_features=   self.n_filters,
            activation=     None,
            bias=           False,
            device=         self.device,
            dtype=          self.dtype,
            initializer=    initializer) if self.in_features != self.n_filters else None

        num_blocks_to_build = 1 if shared_lays else self.n_layers

        self.blocks = [LayBlockCNN(
            n_filters=          self.n_filters,
            padded=             padded,
            kernel_size=        self.kernel_size,
            activation=         activation,
            lay_dropout=        lay_dropout,
            res_dropout=        res_dropout,
            do_ldrt=            do_ldrt,
            ldrt_do_scaled_dns= ldrt_do_scaled_dns,
            ldrt_dns_scale=     ldrt_dns_scale,
            ldrt_drop=          ldrt_drop,
            ldrt_residual=      ldrt_residual,
            ldrt_res_dropout=   ldrt_res_dropout,
            device=             self.device,
            dtype=              self.dtype,
            initializer=        initializer) for _ in range(num_blocks_to_build)]

        for bix,block in enumerate(self.blocks): self.add_module(f'block_{bix}',block)

        if shared_lays and self.n_layers > 1: self.blocks *= self.n_layers

        self.out_ln = torch.nn.LayerNorm(
            normalized_shape=   self.n_filters,
            device=             self.device,
            dtype=              self.dtype)

    # prepares initial history for casual mode, history has shape [.., n_layers, kernel_size-1, n_filters]
    def get_zero_history(self, inp:Optional[TNS]=None) -> TNS:
        block_zero_history = self.blocks[0].get_zero_history(inp)
        bzhs = list(block_zero_history.shape)
        block_zero_history = block_zero_history.view(bzhs[:-2] + [1] + bzhs[-2:]) # add dim
        return block_zero_history.expand(bzhs[:-2] + [self.n_layers] + bzhs[-2:]) # expand

    def forward(
            self,
            inp: TNS,
            history: Optional[TNS]= None) -> DTNS:

        states = []  # here we will store block states to concatenate them finally
        zsL = []

        if self.in_TFdrop_lay:
            inp = self.in_TFdrop_lay(inp)

        if self.projection_lay:
            inp = self.projection_lay(inp)

        output = inp  # for 0 layers case
        histories = torch.split(history, 1, dim=-3) if history is not None else [None]*self.n_layers
        for block,hist in zip(self.blocks, histories):
            if hist is not None: hist = torch.squeeze(hist, dim=-3)
            block_out = block(output, history=hist)
            output = block_out['out']
            if block_out['state'] is not None:
                states.append(torch.unsqueeze(block_out['state'], dim=-3))
            zsL += block_out['zsL']

        output = self.out_ln(output)

        return {
            'out':      output,
            'state':    torch.concat(states,dim=-3) if states else None,
            'zsL':      zsL}


# QKV_linear_projection + QKV_scaled_dot_product_attention + linear_out_projection
class MyMHA(torch.nn.MultiheadAttention):

    # replaces xavier with bert_initializer
    def _reset_parameters(self):

        if self._qkv_same_embed_dim:
            bert_initializer(self.in_proj_weight)
        else:
            bert_initializer(self.q_proj_weight)
            bert_initializer(self.k_proj_weight)
            bert_initializer(self.v_proj_weight)
        bert_initializer(self.out_proj.weight)

        if self.in_proj_bias is not None:
            torch.nn.init.zeros_(self.in_proj_bias)
            torch.nn.init.zeros_(self.out_proj.bias)
        if self.bias_k is not None:
            bert_initializer(self.bias_k)
        if self.bias_v is not None:
            bert_initializer(self.bias_v)


# Block (Layer) of EncTNS (based on torch.nn.modules.transformer.TransformerEncoderLayer)
class LayBlockTNS(torch.nn.Module):

    def __init__(
            self,
            d_model: int=       512,
            nhead: int=         8,
            dns_scale: int=     4,              # up-scale for first dense of two
            dropout: float=     0.1,
            dropout_att: float= 0.0,            # in original (torch.nn..) implementation dropout_att == dropout
            activation: ACT=    torch.nn.ReLU,
            dropout_res: float= 0.0,            # dropout on residual bypass
            device=             None,
            dtype=              None):

        super(LayBlockTNS, self).__init__()

        self.norm1 = torch.nn.LayerNorm(
            normalized_shape=   d_model,
            device=             device,
            dtype=              dtype)

        self.self_attn = MyMHA(
            embed_dim=      d_model,
            num_heads=      nhead,
            dropout=        dropout_att,
            bias=           True,
            add_bias_kv=    False,
            add_zero_attn=  False,
            kdim=           None,
            vdim=           None,
            batch_first=    True,
            device=         device,
            dtype=          dtype)

        self.dropout1 = torch.nn.Dropout(p=dropout) if dropout else None

        self.res1 = LayRES(in_features=d_model, dropout=dropout_res)

        # (LN > Dense or two_with_drop_in_between > drop > RES with drop)
        self.lay_drt = LayBlockDRT(
            in_width=           d_model,
            do_scaled_dns=      True,
            dns_scale=          dns_scale,
            activation=         activation,
            interlay_dropout=   dropout, # INFO: TF implementation has not this dropout
            lay_dropout=        dropout,
            residual=           True,
            res_dropout=        dropout_res,
            device=             device,
            dtype=              dtype,
            initializer=        bert_initializer)

    def forward(
            self,
            inp: TNS,
            task_query: Optional[TNS]=              None,           # forces task-attention mode (TAT)
            inp_mask: Optional[TNS]=                None,
            inp_key_padding_mask: Optional[TNS]=    None) -> DTNS:

        x = inp

        if task_query is None: x = self.norm1(x) # norm first https://arxiv.org/pdf/2002.04745v1.pdf
        else: task_query = self.norm1(task_query) # LN on task_query

        x = self.self_attn(
            query=              x if task_query is None else task_query,
            key=                x,
            value=              x,
            key_padding_mask=   inp_key_padding_mask,
            need_weights=       False,
            attn_mask=          inp_mask)[0]

        if self.dropout1:
            x = self.dropout1(x)

        bypass = inp if task_query is None else task_query
        x = self.res1(inp=x, bypass=bypass)

        return self.lay_drt(x)


# Transformer Encoder (based on torch.nn.modules.transformer.TransformerEncoder), with TAT option and some others
class EncTNS(torch.nn.Module):

    def __init__(
            self,
            num_layers: int=                        6,
            num_layers_TAT: int=                    0,
            initial_TAT_avg: bool=                  True,   # how to prepare first block TAT task_query
            shared_lays: Optional[Tuple[int,...]]=  None,   # tuple defines layers groups with shared variables, e.g.: (2,2,2)
            max_seq_len: Optional[int]=             None,   # when given (int) adds positional embeddings (PE) to seq
            # block params
            d_model: int=                           512,
            nhead: int=                             8,
            dns_scale: int=                         4,
            dropout: float=                         0.1,
            dropout_att: float=                     0.0,
            activation: ACT=                        torch.nn.ReLU,
            dropout_res: float=                     0.0,
            device=                                 None,
            dtype=                                  None):

        super(EncTNS, self).__init__()

        self.d_model = d_model

        # positional embeddings (trainable)
        self.pos_emb = None
        if max_seq_len:
            self.pos_emb = torch.nn.Parameter(
                data=   torch.empty(
                    size=   (max_seq_len, self.d_model),
                    device= device,
                    dtype=  dtype))
            bert_initializer(self.pos_emb)

        self.initial_task_query = None
        if not initial_TAT_avg:
            self.initial_task_query = torch.nn.Parameter(
                data=           torch.empty(self.d_model, device=device, dtype=dtype),
                requires_grad=  False) # TODO: check with experiments
            bert_initializer(self.initial_task_query)

        # manage layers number
        num_layers_to_build = num_layers + num_layers_TAT
        if shared_lays is not None:

            if not sum(shared_lays) == num_layers_to_build:
                raise TorchnessException('Sum of shared layers must be equal to total num of layers (num_layers + num_layers_TAT)')

            num_layers_to_build = len(shared_lays)

        layers = [LayBlockTNS(
            d_model=        self.d_model,
            nhead=          nhead,
            dns_scale=      dns_scale,
            dropout=        dropout,
            dropout_att=    dropout_att,
            activation=     activation,
            dropout_res=    dropout_res,
            device=         device,
            dtype=          dtype,
        ) for _ in range(num_layers_to_build)]
        for lix,lay in enumerate(layers): self.add_module(f'lay_{lix}',lay)

        if shared_lays is not None:
            exp_layers = []
            for lay,mul in zip(layers,shared_lays):
                exp_layers += [lay]*mul
            layers = exp_layers

        self.layers = layers[:num_layers]
        self.layers_TAT = layers[num_layers:]

        self.norm = torch.nn.LayerNorm(
            normalized_shape=   self.d_model,
            device=             device,
            dtype=              dtype)

    # base Transformer encoding
    def _encode(self, inp:TNS, mask:Optional[TNS]=None) -> DTNS:

        output = inp

        """
            it is possible to receive higher than 3-dim input [batch,seq,feats]
            we need to flatten such dim since it is not supported by native torch Transformer:
            "query should be unbatched 2D or batched 3D tensor but received 4-D query tensor <- MultiheadAttention(Module)"
        """
        orig_shape = output.shape
        if len(orig_shape) > 3:
            output = output.view([-1] + list(orig_shape)[-2:])

        # add positional embeddings if needed
        if self.pos_emb is not None:
            output += self.pos_emb[:output.size(-2)]

        zsL = []
        for mod in self.layers:
            block_out = mod(inp=output, inp_mask=mask)
            output = block_out['out']
            zsL += block_out['zsL']

        # pass through task-attention layers
        if self.layers_TAT:

            seq = output


            if self.initial_task_query is not None:
                task_query = self.initial_task_query                # from model
                seq_shape = list(seq.shape)
                seq_shape[-2] = 1
                # TODO: not sure if it will work for batch of sequences, e.g. [128,64,512]
                task_query = task_query.view(seq_shape)
            else:
                task_query = torch.mean(seq, dim=-2, keepdim=True)  # first initial from reduce

            for mod in self.layers_TAT:
                block_out = mod(inp=seq, task_query=task_query, inp_mask=mask)
                task_query = block_out['out']
                zsL += block_out['zsL']
            output = torch.flatten(task_query,-2,-1) # remove seq (-2) dimension of size 1

        output = self.norm(output)

        # eventually roll back to original shape
        if len(orig_shape) > 3:
            output = output.view(orig_shape)

        return {
            'out':  output,
            'zsL':  zsL}

    # pyramidal_encoding
    def _encode_pyramidal(self, inp:TNS, pyramide: Union[Tuple[int],int]) -> DTNS:

        if type(pyramide) is int: pyramide = (pyramide,)

        inp = inp
        zsL = []
        for sl in pyramide:
            in_split = torch.split(inp, sl, dim=0)
            outL = [self._encode(inp=i) for i in in_split]
            inp = [o['out'] for o in outL]
            inp = torch.stack(inp, dim=0)
            zsLL = [o['zsL'] for o in outL]
            zsL += [e for el in zsLL for e in el]
        out = self._encode(inp)
        out['zsL'] += zsL
        return out

    def forward(
            self,
            inp: TNS,
            mask: Optional[TNS]=                        None,
            pyramide: Optional[Union[Tuple[int],int]]=  None) -> DTNS:

        if pyramide:
            if mask is not None:
                raise TorchnessException('mask is not supported for pyramidal encoding')
            return self._encode_pyramidal(inp, pyramide)
        else:
            return self._encode(inp, mask)


from collections import OrderedDict
from pypaq.mpython.mpdecor import proc_wait
import torch
from typing import Optional

from torchness.types import TNS, DTNS



class TorchnessException(Exception):
    pass

# weights initializer from BERT, the only difference is that in torch values are CLAMPED not SAMPLED till in <a,b>
def bert_initializer(*args, std=0.02, **kwargs):
    return torch.nn.init.trunc_normal_(*args, **kwargs, std=std, a=-2*std, b=2*std)


def my_initializer(*args, std=0.02, **kwargs):
    # different layers use different initialization functions:
    # torch Linear % Conv1D uses kaiming_uniform_(weights) & xavier_uniform_(bias)
    # my TF uses trunc_normal_(weights, std=0.02, a==b==2*std) & 0(bias) <- from BERT
    # - kaiming_uniform_ is uniform_ with bound from 2015 paper, (for relu)
    # - xavier_uniform_ is uniform_ whit bound from 2010 paper (for linear / sigmoid)
    # - trunc_normal_ is normal with mean 0 and given std, all values SAMPLED till in <a,b>
    return bert_initializer(*args, **kwargs, std=std)

@proc_wait
def mrg_ckpts(
        ckptA: str,                     # checkpoint A (file name)
        ckptB: Optional[str],           # checkpoint B (file name), for None takes 100% ckptA
        ckptM: str,                     # checkpoint merged (file name)
        ratio: float=           0.5,    # ratio of merge
        noise: float=           0.0     # noise factor, amount of noise added to new value <0.0;1.0>
):
    """
    weighted merge of two checkpoints
    does NOT check for compatibility of two checkpoints, but will crash if those are not compatible
    enclosed with subprocess for better separation of torch objects
    """
    checkpoint_A = torch.load(ckptA)
    checkpoint_B = torch.load(ckptB) if ckptB else checkpoint_A

    cmsd_A = checkpoint_A['model_state_dict']
    cmsd_B = checkpoint_B['model_state_dict']
    cmsd_M = OrderedDict()

    for k in cmsd_A:
        if cmsd_A[k].is_floating_point():
            std_dev = float(torch.std(cmsd_A[k]))
            noise_tensor = torch.zeros_like(cmsd_A[k])
            if std_dev != 0.0: # bias variable case
                my_initializer(noise_tensor, std=std_dev)
            cmsd_M[k] = ratio * cmsd_A[k] + (1 - ratio) * cmsd_B[k] + noise * noise_tensor
        else:
            cmsd_M[k] = cmsd_A[k]

    checkpoint_M = {}
    checkpoint_M.update(checkpoint_A)
    checkpoint_M['model_state_dict'] = cmsd_M

    torch.save(checkpoint_M, ckptM)

# returns base checkpoint information, if given two - checks if B is equal A
def ckpt_nfo(
        ckptA: str,                     # checkpoint A (file name)
        ckptB: Optional[str]=   None,   # checkpoint B (file name)
):
    checkpoint_A = torch.load(ckptA)
    checkpoint_B = torch.load(ckptB) if ckptB else None
    are_equal = True

    cmsd_A = checkpoint_A['model_state_dict']
    cmsd_B = checkpoint_B['model_state_dict'] if checkpoint_B else None

    print(f'Checkpoint has {len(cmsd_A)} tensors, #floats: {sum([cmsd_A[k].numel() for k in cmsd_A])}')
    for k in cmsd_A:
        tns = cmsd_A[k]
        print(f'{k:100} shape: {str(list(tns.shape)):15} {tns.dtype}')
        if cmsd_B:
            if k in cmsd_B:
                if not torch.equal(cmsd_A[k], cmsd_B[k]):
                    print(f' ---> is not equal in second checkpoint')
                    are_equal = False
            else:
                print(f' ---> is not present in second checkpoint')
                are_equal = False
    if checkpoint_B:
        print(f'Checkpoints {"are equal" if are_equal else "are NOT equal"}')


def scaled_cross_entropy(
        labels: TNS,
        scale: TNS,
        logits: Optional[torch.Tensor]= None,
        probs: Optional[torch.Tensor]=  None) -> DTNS:

    if logits is None and probs is None:
        raise TorchnessException('logits OR probs must be given!')

    if probs is None:
        probs = torch.nn.functional.softmax(input=logits, dim=-1)

    prob_label = probs[range(len(labels)), labels] # probability of class from label

    # merge loss for positive and negative scale
    ce = torch.where(
        condition=  scale > 0,
        input=      -torch.log(prob_label),
        other=      -torch.log(1-prob_label))

    return {
        'scaled_cross_entropy': ce * torch.abs(scale),
        'cross_entropy':        ce}


import torch
import unittest

from torchness.layers import LayDense, TF_Dropout, LayConv1D, PositionalEncoding, zeroes


class TestLayers(unittest.TestCase):

    def test_lay_dense(self):
        tns = torch.rand(20)
        print(tns)

        dnsl = LayDense(20,10)
        print(dnsl)
        out = dnsl(tns)
        print(out)
        self.assertTrue(out.size()[0] == 10)
        print(torch.sum(out))
        self.assertTrue(float(torch.sum(out)) >= 0.0)

        dnsl = LayDense(
            in_features=    20,
            out_features=   10,
            activation=     None,
            bias=           False,
            initializer=    torch.nn.init.zeros_)
        out = dnsl(tns)
        print(out)
        print(torch.sum(out))
        self.assertTrue(float(torch.sum(out)) == 0.0)


    def test_tf_dropout(self):
        tns = torch.rand((5,5,5))
        sum_tns = float(torch.sum(tns))
        print(sum_tns)
        drl = TF_Dropout(time_drop=0.3, feat_drop=0.3)
        dropped = drl(tns)
        print(dropped)
        sum_dropped = float(torch.sum(dropped))
        print(sum_dropped)


    def test_tf_dropout_higher_dim(self):
        tns = torch.rand((6,5,12,14))
        sum_tns = float(torch.sum(tns))
        print(sum_tns)
        drl = TF_Dropout(time_drop=0.3, feat_drop=0.6)
        dropped = drl(tns)
        print(dropped.shape)
        sum_dropped = float(torch.sum(dropped))
        print(sum_dropped)


    def test_LayConv1D(self):
        inp = torch.rand(3,6)  # [Channels,SignalSeq]
        print(inp.shape, inp)
        conv_lay = LayConv1D(6,8)
        print(conv_lay)
        out = conv_lay(inp)
        print(out.shape, out)
        self.assertTrue(tuple(out.shape) == (3,8))


    def test_PositionalEncoding(self):

        tns = torch.zeros(12,32,64)
        print(tns)
        print(tns.shape)

        pe = PositionalEncoding(64)
        out = pe(tns)
        print(out)
        print(out.shape)


    def test_zeroes(self):

        tns = torch.rand(5)
        z = zeroes(tns)
        print(z, z.shape)
        self.assertTrue(z.shape[-1]==5)

        tns = torch.rand((3,4,5))
        z = zeroes(tns)
        print(z, z.shape)
        self.assertTrue(z.shape[-1]==5)

        drl = TF_Dropout(time_drop=0.3, feat_drop=0.3)
        dropped = drl(tns)
        print(dropped)
        print(zeroes(dropped))


from setuptools import setup, find_packages

# reads requirements
def get_requirements():
    with open('requirements.txt') as file:
        lines = [l[:-1] if l[-1]=='\n' else l for l in file.readlines()]
        return lines


setup(
    name=               'torchness',
    version=            'v1.1.0',
    url=                'https://github.com/piteren/torchness.git',
    author=             'Piotr Niewinski',
    author_email=       'pioniewinski@gmail.com',
    description=        'PyTorch tools',
    packages=           find_packages(),
    install_requires=   get_requirements(),
    license=            'MIT')


import math
import numpy as np
from typing import List, Optional, Union
from tqdm import tqdm

from torchness.types import TNS, DTNS, INI
from torchness.motorch import Module, MOTorch
from torchness.models.text_embbeder import TextEMB
from torchness.models.simple_feats_classifier import SFeatsCSF



# Simple Text Classification Module, based on Sentence-Transformer
class STextCSF(Module):

    def __init__(
            self,
            st_name: str=                           'all-MiniLM-L6-v2',
            enc_batch_size=                         256,
            in_drop: float=                         0.0,
            mid_width: int=                         30,
            mid_drop: float=                        0.0,
            num_classes: int=                       2,
            class_weights: Optional[List[float]]=   None,
            initializer: INI=                       None,
            dtype=                                  None,
            **kwargs):

        Module.__init__(self, **kwargs)

        self.te_module = TextEMB(
            st_name=        st_name,
            enc_batch_size= enc_batch_size)

        self.csf = SFeatsCSF(
            feats_width=    self.te_module.width,
            in_drop=        in_drop,
            mid_width=      mid_width,
            mid_drop=       mid_drop,
            num_classes=    num_classes,
            class_weights=  class_weights,
            initializer=    initializer,
            dtype=          dtype,
            logger=         self.logger)

    def encode(
            self,
            texts: Union[str, List[str]],
            show_progress_bar=  'auto',
            device=             None) -> DTNS:
        if show_progress_bar == 'auto':
            show_progress_bar = False
            if type(texts) is list and len(texts) > 1000:
                show_progress_bar = True
        embeddings = self.te_module.encode(
            texts=              texts,
            show_progress_bar=  show_progress_bar,
            device=             device)
        return {'embeddings': embeddings}

    def forward(self, feats:TNS) -> DTNS:
        return self.csf(feats)

    def loss(self, feats:TNS, labels:TNS) -> DTNS:
        return self.csf.loss(feats, labels)


class STextCSF_MOTorch(MOTorch):

    def __init__(
            self,
            module_type: Optional[type(STextCSF)]=  STextCSF,
            enc_batch_size=                         128,    # number of lines in batch for embeddings
            fwd_batch_size=                         256,    # number of embeddings in batch for probs
            **kwargs):

        MOTorch.__init__(
            self,
            module_type=    module_type,
            enc_batch_size= enc_batch_size,
            fwd_batch_size= fwd_batch_size,
            **kwargs)

    def get_embeddings(
            self,
            lines: Union[List[str],str],
            show_progress_bar=      'auto') -> np.ndarray:
        if type(lines) is str: lines = [lines]
        self.logger.info(f'{self.name} prepares embeddings for {len(lines)} lines..')
        if show_progress_bar == 'auto':
            show_progress_bar = self.logger.level < 21 and len(lines) > 1000
        out = self.module.encode(
            texts=              lines,
            show_progress_bar=  show_progress_bar,
            device=             self.device) # needs to give device here because of SentenceTransformer bug in encode() #153
        return out['embeddings']

    def get_probs(self, lines:List[str]) -> np.ndarray:

        embs = self.get_embeddings(lines)

        num_splits = math.ceil(embs.shape[0] / self['fwd_batch_size']) # INFO: gives +- batch_size
        featsL = np.array_split(embs,num_splits)

        self.logger.info(f'{self.name} computes probs for {len(featsL)} batches of embeddings')
        iter = tqdm(featsL) if self.logger.level < 21 else featsL
        probsL = [self(feats)['probs'] for feats in iter]
        probs = np.concatenate(probsL)
        self.logger.info(f'> got probs {probs.shape}')

        return probs

    def get_probsL(self, linesL:List[List[str]]) -> List[np.ndarray]:

        lines = []
        for l in linesL:
            lines += l

        probs = self.get_probs(lines)

        acc_lengths = []
        acc = 0
        for l in [len(ls) for ls in linesL]:
            acc_lengths.append(l+acc)
            acc += l
        acc_lengths.pop(-1)

        if acc_lengths: return np.split(probs,acc_lengths)
        else:           return [probs]



import math
import numpy as np
from sentence_transformers import SentenceTransformer
from typing import List, Union
from tqdm import tqdm

from torchness.base_elements import TorchnessException
from torchness.motorch import MOTorch, Module


# Text Embedding Module, based on Sentence-Transformer, prepares embedding for given text (str)
class TextEMB(Module):

    def __init__(
            self,
            st_name: str=   'all-MiniLM-L6-v2',
            enc_batch_size= 256,
            **kwargs):
        Module.__init__(self, **kwargs)
        self.st_name = st_name
        self.st_model = SentenceTransformer(model_name_or_path=st_name)
        self.enc_batch_size = enc_batch_size
        self.logger.info(f'*** TextEMB : {self.st_name} *** initialized, feats width:{self.width} seq length:{self.length}')

    def tokenize(self, texts:List[str]) -> List[List[str]]:
        tokenizer = self.st_model.tokenizer
        return [tokenizer.tokenize(t) for t in texts]

    # original, wrapped version
    def encode(
            self,
            texts: List[str],
            show_progress_bar=  True,
            device=             None,
    ) -> np.ndarray:
        return self.st_model.encode(
            sentences=          texts,
            batch_size=         self.enc_batch_size,
            show_progress_bar=  show_progress_bar,
            device=             device)
    """
    # extracted essence, not working for cuda
    def encode(
            self,
            texts: List[str],
            device=             None,
            show_progress_bar=  True,
    ) -> torch.Tensor:

        # texts should be split in batches
        features = self.st_model.tokenize(texts)
        with torch.no_grad():
            out_features = self.st_model.forward(features)
        #embeddings = out_features['sentence_embedding']
        #embeddings = embeddings.detach().cpu().numpy()

        return out_features['sentence_embedding']
    #"""

    @property
    def width(self) -> int:
        return self.st_model.get_sentence_embedding_dimension()

    @property
    def length(self) -> int:
        return self.st_model.get_max_seq_length()


# is MOTorch for given st_name (SentenceTransformer) based on TextEMB module
class TextEMB_MOTorch(MOTorch):

    def __init__(self, module_type:type(TextEMB)=TextEMB, **kwargs):

        if 'name' not in kwargs:
            if 'st_name' not in kwargs:
                raise TorchnessException('\'st_name\' must be given when MOTorch \'name\' is not given')
            st_name = kwargs['st_name']
            st_name_replaced = st_name.replace('/','__') # replace possible / in st_name since it conflicts with model folder name
            kwargs['name'] = f'TextEMB_MOTorch_{st_name_replaced}'

        MOTorch.__init__(self, module_type=module_type, **kwargs)

    def get_tokens(self, lines: List[str]):
        self.logger.info(f'{self.name} prepares tokens for {len(lines)} lines..')
        return self.module.tokenize(lines)

    def get_embeddings(
            self,
            lines: Union[List[str],str],
            show_progress_bar=  'auto') -> np.ndarray:

        if show_progress_bar == 'auto':
            show_progress_bar = False
            if self.logger.level < 21 and type(lines) is list and len(lines) > 1000:
                show_progress_bar = True

        self.logger.info(f'{self.name} prepares embeddings for {len(lines)} lines..')
        return self.module.encode(
            texts=              lines,
            device=             self.device, # fixes bug of SentenceTransformers.encode() device placement
            show_progress_bar=  show_progress_bar)

    def get_probs(self, lines:List[str]) -> np.ndarray:

        embs = self.get_embeddings(lines)

        num_splits = math.ceil(embs.shape[0] / self['fwd_batch_size']) # INFO: gives +- batch_size
        featsL = np.array_split(embs,num_splits)

        self.logger.info(f'TeXClas_MOTorch computes probs for {len(featsL)} batches of embeddings')
        iter = tqdm(featsL) if self.logger.level < 21 else featsL
        probsL = [self(feats)['probs'].cpu().detach().numpy() for feats in iter]
        probs = np.concatenate(probsL)
        self.logger.info(f'> got probs {probs.shape}')

        return probs

    def get_probsL(self, linesL:List[List[str]]) -> List[np.ndarray]:

        lines = []
        for l in linesL:
            lines += l

        probs = self.get_probs(lines)

        acc_lengths = []
        acc = 0
        for l in [len(ls) for ls in linesL]:
            acc_lengths.append(l+acc)
            acc += l
        acc_lengths.pop(-1)

        if acc_lengths: return np.split(probs,acc_lengths)
        else:           return [probs]

    @property
    def width(self) -> int:
        return self.module.width